
/************************************************************************/
/* Copyright (c) 2016 - 2022 Marnix van den Bos.                        */
/*                                                                      */
/* <marnix@xvan.nl>                                                     */
/*                                                                      */
/* This file is part of XVAN, an Interactive Fiction authoring system.  */
/*                                                                      */
/* XVAN is free software: you can redistribute it and/or modify         */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, version 3.                             */
/*                                                                      */
/* XVAN is distributed in the hope that it will be useful,              */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */
/*                                                                      */
/* You should have received a copy of the GNU General Public License    */
/* along with XVAN.  If not, see <http://www.gnu.org/licenses/>.        */
/*                                                                      */
/************************************************************************/


#include <stdio.h>
#include <string.h>
#include <stdint.h>

#include "keyword.h"
#include "typedefs.h"
#include "debugdev.h"

/* this file contains debug function that I use myself when developing the code */
/* debug functions for the game author are in file debugcmd.c                   */

/* 31oct2018: introduced var text_to_print in init.c that must be malloc-ed */
/* this because I am too lazy to change all the output routines for the  */
/* screen reader. maybe will do this later                               */


/***********/
/* Globals */
/***********/

/*************************/
/* Function declarations */
/*************************/
void    Debug(void);  /* used by XeqDebug() function */

void    PrintStoryInfo(storyInfo*);
void    PrintLocationDirectory(void);
void    PrintObjectDirectory(void);
void    PrintExits(void);
void    PrintCommonAttributes(void);
void    PrintLocalAttributes(void);
void    PrintCommonFlags(void);
void    PrintLocalFlags(void);
void    PrintVerb(verbInfo*);
void    PrintVerbsInMem(void);  /* @!@ */
void    PrintLocation(locationInfo*);
void    PrintObject(objectInfo*);
void    PrintSpecialIds(void);
void    PrintVerbDir(void);
void    PrintWTable(void);
int32_t Test4SpecId(int32_t);
void    PrintParsedInput(parsedInput*);
void    PrintCompActionRec(compActionRec*);
void    PrintUsrActionRec(usrActionRec*);
void    PrintParserActionRec(parserActionRec*);
void    PrintHits(match*, int32_t, int32_t);
void    PrintAllTimers(void);
void    PrintTimer(timerInfo*);
void    PrintSpanTree(spanTree*);
void    PrintRoute(int32_t*);
void    PrintUndoStack(void);
void    PrintChoices(void);  /* @!@ */

/************************/
/* Function definitions */
/************************/

void PrintStoryInfo(storyInfo *info)
{
  printf("\n\n******** Meta data ********\n\n");
  printf("Title: %s\n", info->title);
  printf("Author: %s\n", info->author);
  printf("Organization: %s\n", info->organization);
  printf("Cover text: %s\n", info->cover_text);
  printf("Credits: %s\n", info->credits);
  printf("Version: %s\n", info->version);
  printf("Android market: %s\n", info->android_mkt);
  printf("ios market: %s\n", info->ios_mkt);
  printf("Back image: %s\n", info->backimage);
  printf("Effect: %s\n", info->effect);
  printf("Sidebar: %d\n", info->ui_sidebar);
  printf("Textinput: %d\n", info->ui_textinput);
  printf("Compass: %d\n", info->ui_compass);
  printf("Autolink: %d\n", info->autolink);
  printf("Compiler version: %s\n", info->compiler_version);
  printf("XVAN language: %d\n", info->xvan_language);
  printf("Story language: %d\n", info->story_language);

  printf("\n********************************\n\n");
}


void Debug(void)
{
  PrintSpecialIds();
  PrintLocationDirectory();
  PrintObjectDirectory();
  PrintAllTimers();

}


void PrintLocationDirectory(void)
{
  /* Uses the following Global vars: */
  /*   - loc_dir (type dirInfo*)     */
  /*   - obj_dir (type dirInfo)      */
  /*   - nr_of_locs (type int)       */

  int32_t i, j;
  int32_t offset = 0;
  char    text_to_print[OUTPUT_LINE_LEN];

  PrintString("\n\n**************************\n");
  PrintString("*** LOCATION DIRECTORY ***\n");
  PrintString("**************************\n\n");
  Output(outputbuffer);

  for (i=0; i < nr_of_locs; i++) {
    sprintf(text_to_print, "\nOffset: %ld\n\n", loc_dir[i].offset);
    PrintString(text_to_print);
    Output(outputbuffer);
    sprintf(text_to_print, "Location id: %d\n", i+FIRST_LOCATION_ID);
    PrintString(text_to_print);
    Output(outputbuffer);
    for (j=0; j<loc_dir[i].nr_of_dsys; j++) {
      PrintExtendedSysDescr(&(loc_dir[i].descr[j]));
      PrintString("\n");
      Output(outputbuffer);
    }

    /* There is no containing object or location */

    /* Print the contained objects */
    PrintString("Contained objects:\n");
    Output(outputbuffer);
    for (j=0; j<(loc_dir[i].contained_objs).nr_of_objects; j++) {
      PrintString("  ");
      offset = ((loc_dir[i].contained_objs).object_ids)[j]-FIRST_OBJECT_ID;
      PrintExtendedSysDescr(&(obj_dir[offset].descr[0]));
      sprintf(text_to_print, " (%d)\n", offset+FIRST_OBJECT_ID);
      PrintString(text_to_print);
      Output(outputbuffer);
    }
  }
}


void PrintObjectDirectory(void)
{
  /* Uses the following Global vars: */
  /*   - obj_dir (type dirInfo*)     */
  /*  - loc_dir (type dirInfo*)      */
  /*   - nr_of_objs (type int)       */

  int32_t i, j;
  int32_t offset = 0;
  char    text_to_print[OUTPUT_LINE_LEN];

  PrintString("\n\n************************\n");
  PrintString("*** OBJECT DIRECTORY ***\n");
  PrintString("************************\n\n");
  Output(outputbuffer);

  for (i=0; i < nr_of_objs; i++) {
    sprintf(text_to_print, "\nOffset: %ld\n\n", obj_dir[i].offset);
    PrintString(text_to_print);
    Output(outputbuffer);
    sprintf(text_to_print, "Object id: %d\n", i+FIRST_OBJECT_ID);
    PrintString(text_to_print);
    Output(outputbuffer);
    for (j=0; j<obj_dir[i].nr_of_dsys; j++) {
      PrintWord(obj_dir[i].descr[j].part1.article);
      PrintString(" ");
      PrintExtendedSysDescr(&(obj_dir[i].descr[j]));
      PrintString("\n");
      Output(outputbuffer);
    }

    /* Print the containing object or location */
    PrintString("Contained in: ");
    Output(outputbuffer);
    if (IsLocId((obj_dir[i]).held_by)) {
      offset = (obj_dir[i]).held_by - FIRST_LOCATION_ID;
      PrintExtendedSysDescr(&(loc_dir[offset].descr[0]));         /* print the description, use loc_dir here to find sysDescr as it is a location */
      sprintf(text_to_print, " (%d)\n", obj_dir[i].held_by); /* print the id */
      PrintString(text_to_print);
      Output(outputbuffer);
    }
    else if (IsObjId(obj_dir[i].held_by)) {
      offset = (obj_dir[i]).held_by - FIRST_OBJECT_ID;
      PrintExtendedSysDescr(&(obj_dir[offset].descr[0]));
      sprintf(text_to_print, " (%d)\n", obj_dir[i].held_by);
      PrintString(text_to_print);
      Output(outputbuffer);
    }
    else {
      PrintString("PrintObjectDirectory(): Error: container is not a location or object\n");
      Output(outputbuffer);
      return;
    }

    /* Print the contained objects */
    PrintString("Contained objects:\n");
    Output(outputbuffer);
    for (j=0; j<(obj_dir[i].contained_objs).nr_of_objects; j++) {
      PrintString("  ");
      offset = ((obj_dir[i].contained_objs).object_ids)[j]-FIRST_OBJECT_ID;
      PrintExtendedSysDescr(&(obj_dir[offset].descr[0]));
      sprintf(text_to_print, " (%d)\n", offset+FIRST_OBJECT_ID);
      PrintString(text_to_print);
      Output(outputbuffer);
    }
  }
}


void PrintExits(void)
{
  /* Uses the following Global vars: */
  /*   - exit_data (type *int)       */
  /*- nr_of_directions (type int)    */

  int32_t i, j;

  PrintString("\n\n*****************\n");
  PrintString("*** Exit data ***\n");
  PrintString("*****************\n\n");
  Output(outputbuffer);

  i = 0;
  while (i<nr_of_locs) {
    PrintId(i+FIRST_LOCATION_ID);
    PrintString(":\n");
    Output(outputbuffer);
    for (j=0; j<nr_of_directions; j++) {
      if (exit_data[i*nr_of_directions+j] != NO_ID) {
        PrintId(j+first_direction_id);
        PrintString("-> ");
        PrintId(exit_data[i*nr_of_directions+j]);
        PrintString("\n");
        Output(outputbuffer);
      }
    }
    PrintString("---------------\n");
    Output(outputbuffer);
    i++;
  }
}


void PrintCommonAttributes(void)
{
  int32_t i;
  int32_t total_attrs; /* Total number of common attributes for all locations */
  char    text_to_print[OUTPUT_LINE_LEN];

  PrintString("\n\n*****************************************************************\n");
  PrintString("*** Common Location Attributes (id, type, value, value owner) ***\n"); /* 10march2017 */
  PrintString("*****************************************************************\n\n");
  Output(outputbuffer);

  total_attrs = nr_of_cattrs*nr_of_locs;

  for (i=0; i<total_attrs; i++) {
    sprintf(text_to_print, "%d, %d, %d %d\n", i+FIRST_COMMON_ATTR_ID, c_loc_attrs[i].type,
            c_loc_attrs[i].value, c_obj_attrs[i].value_owner);
    PrintString(text_to_print);
    Output(outputbuffer);
  }

  PrintString("\n\n\n************************************************************\n");
  PrintString("*** Common Object Attributes (type, value, value owner) ***\n"); /* 10march2017 */
  PrintString("************************************************************\n\n");
  Output(outputbuffer);

  total_attrs = nr_of_cattrs*nr_of_objs;

  for (i=0; i<total_attrs; i++) {
    sprintf(text_to_print, "%d, %d, %d, %d\n", i+FIRST_COMMON_ATTR_ID, c_obj_attrs[i].type,
            c_obj_attrs[i].value, c_obj_attrs[i].value_owner); /* 10march2017 */
    PrintString(text_to_print);
    Output(outputbuffer);
  }
  PrintString("\n");
  Output(outputbuffer);
}


void PrintLocalAttributes(void)
{
  int32_t i;
  char    text_to_print[OUTPUT_LINE_LEN];

  /* Total number of local attributes is stored in global variable nr_of_lattrs */

  PrintString("\n\n***************************************************\n");
  PrintString("*** Local Attributes (id, type, value, value owner) ***\n"); /* 10march2017 */
  PrintString("*******************************************************\n\n");
  Output(outputbuffer);

  for (i=0; i<nr_of_lattrs; i++) {
    sprintf(text_to_print, "%d, %d, %d %d\n", i+FIRST_LOCAL_ATTR_ID, local_attrs[i].type, /* 10march2017 */
            local_attrs[i].value, local_attrs[i].value_owner);
    PrintString(text_to_print);
    Output(outputbuffer);
  }
  PrintString("\n");
  Output(outputbuffer);
}


void PrintCommonFlags(void)
{
  int32_t i;
  int32_t com_loc_flags_string_len = 0;
  int32_t com_obj_flags_string_len = 0;
  char    text_to_print[OUTPUT_LINE_LEN];

  /* calculate the length of the common flag strings */
  com_loc_flags_string_len = ((nr_of_cflags*nr_of_locs)/WORD_LEN)+1;
  com_obj_flags_string_len = ((nr_of_cflags*nr_of_objs)/WORD_LEN)+1;

  PrintString("\n\n************************************\n");
  PrintString("*** Common location flags in hex ***\n");
  PrintString("************************************\n\n");
  Output(outputbuffer);

  for (i=0; i<com_loc_flags_string_len; i++) {
    sprintf(text_to_print, "    %x", com_loc_flags[i]);
    PrintString(text_to_print);
    Output(outputbuffer);
  }
  PrintString("\n");
  Output(outputbuffer);

  PrintString("\n\n**********************************\n");
  PrintString("*** Common object flags in hex ***\n");
  PrintString("**********************************\n\n");
  Output(outputbuffer);

  for (i=0; i<com_obj_flags_string_len; i++) {
    sprintf(text_to_print, "    %x", com_obj_flags[i]);
    PrintString(text_to_print);
    Output(outputbuffer);
  }
  PrintString("\n");
  Output(outputbuffer);
}


void PrintLocalFlags(void)
{
  int32_t i;
  char    text_to_print[OUTPUT_LINE_LEN];

  /* Global var loc_flags_string_len is used here */

  PrintString("\n\n**************************\n");
  PrintString("*** Local flags in hex ***\n");
  PrintString("**************************\n\n");
  Output(outputbuffer);

  for (i=0; i<loc_flags_string_len; i++) {
    sprintf(text_to_print, "    %x", local_flags[i]);
    PrintString(text_to_print);
    Output(outputbuffer);
  }
  PrintString("\n");
  Output(outputbuffer);
}


void PrintLocation(locationInfo *loc)
{
  descrInfo   *descr;
  triggerInfo *trigger;
  int32_t i = 0;
  char        text_to_print[OUTPUT_LINE_LEN];

  PrintString("\n\n**** Location Info ****\n\n");
  PrintString("\n\n Descriptions:\n");
  Output(outputbuffer);

  descr = loc->descriptions;
  while (descr != NULL) {
    sprintf(text_to_print, "\n  Description %d: %s\n", descr->id, descr->text);
    PrintString(text_to_print);
    Output(outputbuffer);
    descr = descr->next;
  }

  if (loc->triggers == NULL) {
    PrintString("No triggers\n");
    Output(outputbuffer);
  }
  else {
    PrintString("\n\n Triggers\n");
    Output(outputbuffer);
  }

  trigger = loc->triggers;
  while (trigger != NULL) {
    i = 0;
    sprintf(text_to_print, "\n  Trigger %d: \n", trigger->id);
    PrintString(text_to_print);
    Output(outputbuffer);
    while (trigger->code[i] != END_OF_CODE) {
      sprintf(text_to_print, "   %d\n", trigger->code[i++]);
      PrintString(text_to_print);
      Output(outputbuffer);
    }
    sprintf(text_to_print, "   %d\n", trigger->code[i]);     /* end of code */
    PrintString(text_to_print);
    Output(outputbuffer);
    trigger = trigger->next;
  }

  PrintString("\n***********************\n");
  Output(outputbuffer);
}


void PrintVerb(verbInfo *verb)
{
  compActionRec *arec;
  int32_t       *code;
  int32_t       i = 0;
  char          text_to_print[OUTPUT_LINE_LEN];

  PrintString("\n\n**** Verb Info ****\n\n");
  PrintString(" Action records:\n");
  Output(outputbuffer);

  arec = verb->action_rec;
  while (arec != NULL) {
    PrintCompActionRec(arec);
    arec = arec->next;
  }

  PrintString("Trigger code:\n");
  Output(outputbuffer);

  code = verb->code;
  while (code[i] != END_OF_CODE) {
    sprintf(text_to_print, "%d ", code[i++]);
    PrintString(text_to_print);
    Output(outputbuffer);
  }

  PrintString("\n\n***********************\n");
  Output(outputbuffer);
}


void PrintVerbsInMem(void)  /* @!@ */
{
  int  i = 0;
  char text_to_print[OUTPUT_LINE_LEN];

  PrintString("Following verbs are currently loaded in memory:\n\n");
  for (i=0; i<nr_of_verbs; i++) {
    if (InMem(i+FIRST_VERB_ID)) {
      /* verb is in memory */
      sprintf(text_to_print, "%d: ", i+FIRST_VERB_ID);
      PrintString(text_to_print);
      PrintId(i+FIRST_VERB_ID);
      PrintString("\n");
      Output(outputbuffer);
    }
  }
}


void PrintObject(objectInfo *obj)
{
  descrInfo   *descr;
  triggerInfo *trigger;
  int32_t i = 0;
  char        text_to_print[OUTPUT_LINE_LEN];

  PrintString("\n\n**** Object Info ****\n\n");
  PrintString("\n\n Descriptions:\n");
  Output(outputbuffer);

  descr = obj->descriptions;
  while (descr != NULL) {
    sprintf(text_to_print, "\n  Description %d: %s\n", descr->id, descr->text);
    PrintString(text_to_print);
    Output(outputbuffer);
    descr = descr->next;
  }

  if (obj->triggers == NULL) {
    PrintString("No triggers\n");
    Output(outputbuffer);
  }
  else {
    PrintString("\n\n Triggers\n");
    Output(outputbuffer);
  }

  trigger = obj->triggers;
  while (trigger != NULL) {
    i = 0;
    sprintf(text_to_print, "\n  Trigger %d: \n", trigger->id);
    PrintString(text_to_print);
    Output(outputbuffer);
    while (trigger->code[i] != END_OF_CODE) {
      sprintf(text_to_print, "%d ", trigger->code[i++]);
      PrintString(text_to_print);
      Output(outputbuffer);
    }
    sprintf(text_to_print, "%d\n", trigger->code[i]);     /* end of code */
    PrintString(text_to_print);
    Output(outputbuffer);
    trigger = trigger->next;
  }

  PrintString("\n***********************\n");
  Output(outputbuffer);
}


void PrintSpecialIds(void)
{
  PrintString("*** Special Ids ***\n");

  PrintString("Current location: ");
  Output(outputbuffer);
  PrintId(curr_loc);
  PrintString("\n");
  Output(outputbuffer);

  PrintString("Actor:            ");
  Output(outputbuffer);
  PrintId(actor);
  PrintString("\n");
  Output(outputbuffer);

  PrintString("subject:          ");
  Output(outputbuffer);
  PrintId(subject);
  PrintString("\n");
  Output(outputbuffer);

  PrintString("specifier:        ");
  Output(outputbuffer);
  PrintId(specifier);
  PrintString("\n");
  Output(outputbuffer);

  PrintString("action:           ");
  Output(outputbuffer);
  PrintWord(action);
  PrintString("\n");
  Output(outputbuffer);

  PrintString("direction:        ");
  Output(outputbuffer);
  PrintWord(direction);
  PrintString("\n");
  Output(outputbuffer);

  PrintString("preposition:      ");
  Output(outputbuffer);
  PrintWord(prepos);
  PrintString("\n");
  Output(outputbuffer);

  PrintString("This:             ");
  Output(outputbuffer);
  PrintId(THIS);
  PrintString("\n");
  Output(outputbuffer);

  PrintString("Value:            ");
  Output(outputbuffer);
  PrintId(VALUE);
  PrintString("\n");
  Output(outputbuffer);

  PrintString("Ordinal           ");
  Output(outputbuffer);
  PrintId(ORDINAL);
  PrintString("\n");
  Output(outputbuffer);

  PrintString("*******************\n");
  Output(outputbuffer);
}

void PrintVerbDir(void)
{
  int i  = 0;

  printf("\n\n******** VERB DIRECTORY ********\n\n");
  printf("nr_of_verbs: %d\n", nr_of_verbs);
  for(i=0; i<nr_of_verbs; i++) {
    printf("\nverb id: %d (%d)\n", i+FIRST_VERB_ID, i);
    printf("scope: %d\n", verb_dir[i].scope);
    printf("offset: %ld\n", verb_dir[i].offset);
  }
  printf("\n\n********************************\n\n");
}


void PrintWTable(void)
{
  int32_t i = 0;
  char    text_to_print[OUTPUT_LINE_LEN];

  /* print the word_table */
  PrintString("\n\n*** word_table ***\n\n");
  PrintString("word print_word\n\n");
  Output(outputbuffer);

  for (i=0; i<nr_of_words; i++) {
    sprintf(text_to_print, "%d %s, %s\n", word_table[i].id, word_table[i].word, word_table[i].print_word);
    PrintString(text_to_print);
    Output(outputbuffer);
  }
  PrintString("\n\n");
  Output(outputbuffer);
}


void PrintParsedInput(parsedInput *parsed_input)
{
  int32_t i = 0;
  int32_t j = 0;
  char    text_to_print[OUTPUT_LINE_LEN];

  PrintString("**** Parsed Input ****\n");
  Output(outputbuffer);

  sprintf(text_to_print, "Action1: ");
  PrintWord(parsed_input->action1);
  PrintString("\n");
  Output(outputbuffer);

  PrintString("Action2: ");
  PrintWord(parsed_input->action2);
  PrintString("\n");
  Output(outputbuffer);

  PrintString("\nActor - Part 1:\n");
  sprintf(text_to_print, " %d adjectives\n ", parsed_input->actor.part1.nr_of_adjectives);
  PrintString(text_to_print);
  Output(outputbuffer);

  for (i=0; i<parsed_input->actor.part1.nr_of_adjectives; i++) {
    PrintWord(parsed_input->actor.part1.adjectives[i]);
    Output(outputbuffer);
  }

  PrintString("\n noun: ");
  PrintWord(parsed_input->actor.part1.noun);
  Output(outputbuffer);

  PrintString("\n connecting preposition: ");
  PrintWord(parsed_input->actor.connect_prepos);
  Output(outputbuffer);

  PrintString("\nActor - Part 2:\n");
  sprintf(text_to_print, " %d adjectives\n ", parsed_input->actor.part2.nr_of_adjectives);
  PrintString(text_to_print);
  Output(outputbuffer);

  for (i=0; i<parsed_input->actor.part2.nr_of_adjectives; i++) {
    PrintWord(parsed_input->actor.part2.adjectives[i]);
    Output(outputbuffer);
  }

  PrintString("\n noun: ");
  PrintWord(parsed_input->actor.part2.noun);
  Output(outputbuffer);

  PrintString("\nAdverbs:\n");

  sprintf(text_to_print, "%d adverbs\n ", parsed_input->adverbs.nr_of_adverbs);
  PrintString(text_to_print);
  Output(outputbuffer);

  for (i=0;i<parsed_input->adverbs.nr_of_adverbs;i++) {
    PrintWord(parsed_input->adverbs.adverb[i]);
    Output(outputbuffer);
  }

  PrintString("\nQ-word: ");
  PrintWord(parsed_input->q_word);
  Output(outputbuffer);

  PrintString("\nDirection: ");
  PrintWord(parsed_input->direction);
  Output(outputbuffer);

  PrintString("\nSubject(s):\n");
  Output(outputbuffer);

  for (i=0; i<MAX_SUBJECTS; i++) {
    sprintf(text_to_print, "\nSubject %d - part 1\n ", i);
    PrintString(text_to_print);
    Output(outputbuffer);
    sprintf(text_to_print, " %d adjectives\n ", parsed_input->subject[i].part1.nr_of_adjectives);
    PrintString(text_to_print);
    Output(outputbuffer);
    for (j=0; j<parsed_input->subject[i].part1.nr_of_adjectives; j++) {
      PrintWord(parsed_input->subject[i].part1.adjectives[j]);
      Output(outputbuffer);
    }
    PrintString("\n noun: ");
    PrintWord(parsed_input->subject[i].part1.noun);
    Output(outputbuffer);

    sprintf(text_to_print, "\n connecting preposition: ");
    PrintWord(parsed_input->subject[i].connect_prepos);
    Output(outputbuffer);

    sprintf(text_to_print, "\nSubject %d - part 2\n ", i);
    PrintString(text_to_print);
    Output(outputbuffer);
    sprintf(text_to_print, " %d adjectives\n ", parsed_input->subject[i].part2.nr_of_adjectives);
    PrintString(text_to_print);
    Output(outputbuffer);
    for (j=0; j<parsed_input->subject[i].part2.nr_of_adjectives; j++) {
      PrintWord(parsed_input->subject[i].part2.adjectives[j]);
      Output(outputbuffer);
    }
    PrintString("\n noun: ");
    PrintWord(parsed_input->subject[i].part2.noun);
    PrintString("\n");
    Output(outputbuffer);
  }

  PrintString("\nSpecifier - part 1:\n");
  sprintf(text_to_print, " %d adjectives\n ", parsed_input->specifier.part1.nr_of_adjectives);
  PrintString(text_to_print);
  Output(outputbuffer);
  for (i=0; i<parsed_input->specifier.part1.nr_of_adjectives; i++) {
    PrintWord(parsed_input->specifier.part1.adjectives[i]);
    Output(outputbuffer);
  }
  PrintString("\n noun: ");
  PrintWord(parsed_input->specifier.part1.noun);
  Output(outputbuffer);

  PrintString("\n connecting preposition: ");
  PrintWord(parsed_input->specifier.connect_prepos);
  Output(outputbuffer);

  PrintString("\nSpecifier - part 2:\n");
  sprintf(text_to_print, " %d adjectives\n ", parsed_input->specifier.part2.nr_of_adjectives);
  PrintString(text_to_print);
  Output(outputbuffer);
  for (i=0; i<parsed_input->specifier.part2.nr_of_adjectives; i++) {
    PrintWord(parsed_input->specifier.part2.adjectives[i]);
    Output(outputbuffer);
  }
  PrintString("\n noun: ");
  PrintWord(parsed_input->specifier.part2.noun);
  Output(outputbuffer);

  PrintString("\nPrepositions:\n");
  sprintf(text_to_print, " %d prepositions\n ", parsed_input->prepositions.nr_of_prepositions);
  PrintString(text_to_print);
  Output(outputbuffer);
  for (i=0; i<parsed_input->prepositions.nr_of_prepositions; i++) {
    PrintWord(parsed_input->prepositions.preposition[i]);
    Output(outputbuffer);
  }

  PrintString("\nValue:\n");
  sprintf(text_to_print, " %d\n ", parsed_input->value);
  PrintString(text_to_print);
  Output(outputbuffer);

  PrintString("\nOrdinal:\n");
  sprintf(text_to_print, " %d\n ", parsed_input->ordinal);
  PrintString(text_to_print);
  Output(outputbuffer);

  PrintString("\n**********************\n");
  Output(outputbuffer);
}


int32_t Test4SpecId(int32_t id)
{
  /* This functin test whether id has one of the following values: */
  /* NO_ID, CURR_LOC, SUBJECT, ACTOR, SPECIFIER, and prints a      */
  /* corresponding message.                                        */

  switch (id) {
    case NO_ID:
      PrintString("<none>");
      Output(outputbuffer);
      return(OK);
    case CURR_LOC:
      PrintString("<the current location>\n");
      Output(outputbuffer);
      return(OK);
    case SUBJECT:
      PrintString("<the current subject>\n");
      Output(outputbuffer);
      return(OK);
    case ACTOR:
      PrintString("<the current actor>\n");
      Output(outputbuffer);
      return(OK);
    case SPECIFIER:
      PrintString("<the current specifier>\n");
      Output(outputbuffer);
      return(OK);
    default:
      /* Not a special id. */
      return(ERROR);
  } /* switch */
  /* No OK here, makes compiler happy. */
}


void PrintCompActionRec(compActionRec *action_rec)
{
  int32_t i = 0;
  char    text_to_print[OUTPUT_LINE_LEN];

  PrintString("**** Computer Action Record ****\n");
  Output(outputbuffer);

  PrintString("\nAction1: ");
  Output(outputbuffer);
  if (action_rec->action1 == NO_ID) {
    PrintString("<none>\n");
    Output(outputbuffer);
  }
  else
    if (action_rec->action1 == DEFAULT) {
      PrintString("DEFAULT");
      Output(outputbuffer);
    }
    else
      PrintWord(action_rec->action1);
  PrintString("\n");
  Output(outputbuffer);

  PrintString("\nAction2: ");
  Output(outputbuffer);
  if (action_rec->action2 == NO_ID) {
    PrintString("<none>\n");
    Output(outputbuffer);
  }
  else
    if (action_rec->action2 == DEFAULT) {
      PrintString("DEFAULT");
      Output(outputbuffer);
    }
    else
      PrintWord(action_rec->action2);
  PrintString("\n");
  Output(outputbuffer);

  PrintString("\nActor: ");
  Output(outputbuffer);
  /* Test for special ids. */
  if (!Test4SpecId(action_rec->actor))
    /* Not a special id. */
      PrintId(action_rec->actor);
  PrintString("\n");
  Output(outputbuffer);

  PrintString("\nAdverbs:\n");
  Output(outputbuffer);
  sprintf(text_to_print, "  %d adverbs\n ", action_rec->adverbs.nr_of_adverbs);
  PrintString(text_to_print);
  Output(outputbuffer);
  for (i=0;i<action_rec->adverbs.nr_of_adverbs;i++) {
    PrintWord(action_rec->adverbs.adverb[i]);
    PrintString(" ");
    Output(outputbuffer);
  }
  PrintString("\n");
  Output(outputbuffer);

  PrintString("\nQ-word: ");
  Output(outputbuffer);
  if (action_rec->q_word == NO_ID) {
    PrintString("<none>");
    Output(outputbuffer);
  }
  else
    PrintWord(action_rec->q_word);
  PrintString("\n");
  Output(outputbuffer);

  sprintf(text_to_print, "\nDirection: ");
  Output(outputbuffer);
  if (action_rec->direction == NO_ID) {
    PrintString("<none>");
    Output(outputbuffer);
  }
  else {
    sprintf(text_to_print, "%d\n", action_rec->direction);
    PrintString(text_to_print);
    Output(outputbuffer);
  }

  sprintf(text_to_print, "\nSubject: %d ", action_rec->subject);
  PrintString(text_to_print);
  Output(outputbuffer);
  /* Test for special ids. */
  if (!Test4SpecId(action_rec->subject))
    /* Not a special id. */
    PrintId(action_rec->subject);
  PrintString("\n");
  Output(outputbuffer);

  PrintString("\nSpecifier: ");
  Output(outputbuffer);
  /* Test for special ids. */
  if (!Test4SpecId(action_rec->specifier))
    /* Not a special id. */
    PrintId(action_rec->specifier);
  PrintString("\n");
  Output(outputbuffer);

  PrintString("\nPrepositions:\n");
  Output(outputbuffer);
  sprintf(text_to_print, " %d prepositions\n ", action_rec->prepositions.nr_of_prepositions);
  PrintString(text_to_print);
  Output(outputbuffer);
  for (i=0; i<action_rec->prepositions.nr_of_prepositions; i++) {
    PrintWord(action_rec->prepositions.preposition[i]);
    PrintString(" ");
    Output(outputbuffer);
  }
  PrintString("\n");
  Output(outputbuffer);

  sprintf(text_to_print, "\nValue: %d\n", action_rec->value);
  PrintString(text_to_print);
  Output(outputbuffer);

  sprintf(text_to_print, "\nOrdinal: %d\n", action_rec->ordinal);
  PrintString(text_to_print);
  Output(outputbuffer);

  sprintf(text_to_print, "Trigger: %d %d\n", action_rec->execute[0], action_rec->execute[1]);
  PrintString(text_to_print);
  Output(outputbuffer);
  PrintString("\n********************\n");
  Output(outputbuffer);
}


void PrintUsrActionRec(usrActionRec *action_rec)
{
  int32_t i = 0;
  char    text_to_print[OUTPUT_LINE_LEN];

  PrintString("**** User Action Record ****\n");
  Output(outputbuffer);

  PrintString("\nAction1: ");
  Output(outputbuffer);
  if (action_rec->action1 == NO_ID) {
    PrintString("<none>\n");
    Output(outputbuffer);
  }
  else
    if (action_rec->action1 == DEFAULT) {
      PrintString("DEFAULT");
      Output(outputbuffer);
    }
    else {
      printf("%d", action_rec->action1);  /* @!@ */
      PrintWord(action_rec->action1);
    }
  PrintString("\n");
  Output(outputbuffer);

    PrintString("\nAction2: ");
    Output(outputbuffer);
  if (action_rec->action2 == NO_ID) {
    PrintString("<none>\n");
    Output(outputbuffer);
  }
  else
    if (action_rec->action2 == DEFAULT) {
      PrintString("DEFAULT");
      Output(outputbuffer);
    }
    else
      PrintWord(action_rec->action2);
  PrintString("\n");
  Output(outputbuffer);

  PrintString("\nActor: ");
  Output(outputbuffer);
  /* Test for special ids. */
  if (!Test4SpecId(action_rec->actor))
    /* Not a special id. */
      PrintId(action_rec->actor);
  PrintString("\n");
  Output(outputbuffer);

  PrintString("\nAdverbs:\n");
  Output(outputbuffer);
  sprintf(text_to_print, "  %d adverbs\n ", action_rec->adverbs.nr_of_adverbs);
  PrintString(text_to_print);
  Output(outputbuffer);
  for (i=0;i<action_rec->adverbs.nr_of_adverbs;i++) {
    PrintWord(action_rec->adverbs.adverb[i]);
    PrintString(" ");
    Output(outputbuffer);
  }
  sprintf(text_to_print, "\n");
  Output(outputbuffer);

  PrintString("\nQ-word: ");
  Output(outputbuffer);
  if (action_rec->q_word == NO_ID) {
    PrintString("<none>");
    Output(outputbuffer);
  }
  else
    PrintWord(action_rec->q_word);
  PrintString("\n");
  Output(outputbuffer);

  PrintString("\nDirection: ");
  Output(outputbuffer);
  if (action_rec->direction == NO_ID) {
    PrintString("<none>");
    Output(outputbuffer);
  }
  else {
    sprintf(text_to_print, "%d\n", action_rec->direction);
    PrintString(text_to_print);
    Output(outputbuffer);
  }

  for (i=0; i<MAX_SUBJECTS; i++) {
    sprintf(text_to_print, "\nSubject: %d ", action_rec->subject[i]);
    PrintString(text_to_print);
    Output(outputbuffer);
    /* Test for special ids. */
    if (!Test4SpecId(action_rec->subject[i]))
      /* Not a special id. */
      PrintId(action_rec->subject[i]);
    PrintString("\n");
    Output(outputbuffer);
  }

  PrintString("\nSpecifier: ");
  Output(outputbuffer);
  /* Test for special ids. */
  if (!Test4SpecId(action_rec->specifier))
    /* Not a special id. */
    PrintId(action_rec->specifier);
  PrintString("\n");
  Output(outputbuffer);

  PrintString("\nPrepositions:\n");
  Output(outputbuffer);
  sprintf(text_to_print, " %d prepositions: ", action_rec->prepositions.nr_of_prepositions);
  PrintString(text_to_print);
  Output(outputbuffer);
  for (i=0; i<action_rec->prepositions.nr_of_prepositions; i++) {
    PrintWord(action_rec->prepositions.preposition[i]);
    PrintString(" ");
    Output(outputbuffer);
  }

  sprintf(text_to_print, "\nValue: %d\n ", action_rec->value);
  PrintString(text_to_print);
  Output(outputbuffer);

  sprintf(text_to_print, "Ordinal: %d\n ", action_rec->ordinal);
  PrintString(text_to_print);
  Output(outputbuffer);

  PrintString("\n********************\n");
  Output(outputbuffer);
}


void PrintParserActionRec(parserActionRec *action_rec)
{
  int32_t i = 0;
  char    text_to_print[OUTPUT_LINE_LEN];

  while (action_rec != NULL) {
    PrintString("**** Parser Action Record ****\n");
    Output(outputbuffer);

    PrintString("\nAction1: ");
    Output(outputbuffer);
    if (action_rec->action1 == NO_ID) {
      PrintString("<none>\n");
      Output(outputbuffer);
    }
    else
      if (action_rec->action1 == DEFAULT) {
        PrintString("DEFAULT");
        Output(outputbuffer);
      }
      else
        PrintWord(action_rec->action1);
    PrintString("\n");
    Output(outputbuffer);

    PrintString("\nAction2: ");
    Output(outputbuffer);
    if (action_rec->action2 == NO_ID) {
      PrintString("<none>\n");
      Output(outputbuffer);
    }
    else
      if (action_rec->action2 == DEFAULT) {
        PrintString("DEFAULT");
        Output(outputbuffer);
      }
      else
        PrintWord(action_rec->action2);
    PrintString("\n");
    Output(outputbuffer);

    PrintString("\nActor: ");
    Output(outputbuffer);
    /* Test for special ids. */
    if (!Test4SpecId(action_rec->actor))
      /* Not a special id. */
        PrintId(action_rec->actor);
    PrintString("\n");
    Output(outputbuffer);

    PrintString("\nAdverbs:\n");
    Output(outputbuffer);
    sprintf(text_to_print, "  %d adverbs\n ", action_rec->adverbs.nr_of_adverbs);
    PrintString(text_to_print);
    Output(outputbuffer);
    for (i=0;i<action_rec->adverbs.nr_of_adverbs;i++) {
      PrintWord(action_rec->adverbs.adverb[i]);
      PrintString(" ");
      Output(outputbuffer);
    }
    PrintString("\n");
    Output(outputbuffer);

    PrintString("\nQ-word: ");
    Output(outputbuffer);
    if (action_rec->q_word == NO_ID) {
      PrintString("<none>");
      Output(outputbuffer);
    }
    else
      PrintWord(action_rec->q_word);
    PrintString("\n");
    Output(outputbuffer);

    PrintString("\nDirection: ");
    Output(outputbuffer);
    if (action_rec->direction == NO_ID) {
      PrintString("<none>");
      Output(outputbuffer);
    }
    else {
      sprintf(text_to_print, "%d\n", action_rec->direction);
      PrintString(outputbuffer);
      Output(outputbuffer);
    }

    PrintString("\nSubject: ");
    Output(outputbuffer);
    /* Test for special ids. */
    if (!Test4SpecId(action_rec->subject)) {
      /* Not a special id. */
      PrintId(action_rec->subject);
      Output(outputbuffer);
    }
    else {
      sprintf(text_to_print, "special id: %d", action_rec->subject);
      PrintString(text_to_print);
      Output(outputbuffer);
    }
    PrintString("\n");
    Output(outputbuffer);

    PrintString("\nSpecifier: ");
    Output(outputbuffer);
    /* Test for special ids. */
    if (!Test4SpecId(action_rec->specifier))
      /* Not a special id. */
      PrintId(action_rec->specifier);
    PrintString("\n");
    Output(outputbuffer);

    PrintString("\nPrepositions:\n");
    Output(outputbuffer);
    sprintf(text_to_print, " %d prepositions\n ", action_rec->prepositions.nr_of_prepositions);
    PrintString(text_to_print);
    Output(outputbuffer);
    for (i=0; i<action_rec->prepositions.nr_of_prepositions; i++) {
      PrintWord(action_rec->prepositions.preposition[i]);
      PrintString(" ");
      Output(outputbuffer);
    }

    sprintf(text_to_print, "\nValue: %d\n ", action_rec->value);
    PrintString(text_to_print);
    Output(outputbuffer);

    sprintf(text_to_print, "Ordinal: %d\n ", action_rec->ordinal);
    PrintString(text_to_print);
    Output(outputbuffer);

    sprintf(text_to_print, "Score: %d\n ", action_rec->score);
    PrintString(text_to_print);
    Output(outputbuffer);

    PrintString("\n********************\n");
    Output(outputbuffer);

    action_rec = action_rec->next;
  }
}


void PrintHits(match *hits, int32_t scope, int32_t target)
{
  int i = 0;

  PrintString("\n********* Hits *********\n\n");
  PrintString("Hits for ");
  Output(outputbuffer);
  switch (scope) {
    case ALL_LOCS:
      PrintString("All locations, ");
      break;
    case ACTOR_ONLY:
      PrintString("Actor only, ");
      break;
    case CURR_LOC_ONLY:
      PrintString("Current location, ");
      break;
    default:
      PrintString("<unknown scope>, ");
      break;
  }
  Output(outputbuffer);

  switch (target) {
    case ACTOR:
      PrintString("Actor.\n");
      break;
    case SUBJECT:
      PrintString("Subject.\n");
      break;
    case SPECIFIER:
      PrintString("Specifier.\n");
      break;
    default:
      PrintString("<unknown target>\n");
      break;
  }
  Output(outputbuffer);

  for (i=0; i<hits->nr_of_hits; i++) {
    PrintId(hits->matched_objs[i]);
    PrintString("\n");
    Output(outputbuffer);
  }

  PrintString("\n");
  PrintString("\n************************\n\n");
  Output(outputbuffer);
}


void PrintAllTimers(void)
{
  int32_t i = 0;

  for (i=0; i<nr_of_timers; i++) {
    PrintTimer(timers+i);
    Output(outputbuffer);
  }
  PrintString("\n\n");
  Output(outputbuffer);
}


void PrintTimer(timerInfo *t)
{
  char text_to_print[OUTPUT_LINE_LEN];

  PrintString("**********Timer info**********\n\n");
  Output(outputbuffer);

  sprintf(text_to_print, "value: %d\n", t->value);
  PrintString(text_to_print);
  Output(outputbuffer);
  sprintf(text_to_print, "step: %d\n", t->step);
  PrintString(text_to_print);
  Output(outputbuffer);
  sprintf(text_to_print, "interval: %d\n", t->interval);
  PrintString(text_to_print);
  Output(outputbuffer);
  sprintf(text_to_print, "update: %d\n", t->update);
  PrintString(text_to_print);
  Output(outputbuffer);
  sprintf(text_to_print, "direction: %d\n", t->direction);
  PrintString(text_to_print);
  Output(outputbuffer);
  sprintf(text_to_print, "state: %d\n", t->state);
  PrintString(text_to_print);
  Output(outputbuffer);
  sprintf(text_to_print, "trigger_at: %d\n", t->trigger_at);
  PrintString(text_to_print);
  Output(outputbuffer);
  sprintf(text_to_print, "trigger_spec: %d\n", t->trigger_spec);
  PrintString(text_to_print);
  Output(outputbuffer);
  sprintf(text_to_print, "owner execute: %d\n", t->execute[0]);
  PrintString(text_to_print);
  Output(outputbuffer);
  sprintf(text_to_print, "id execute: %d\n", t->execute[1]);
  PrintString(text_to_print);
  Output(outputbuffer);

  sprintf(text_to_print, "\n******************************\n");
  Output(outputbuffer);
}


void PrintSpanTree(spanTree *tree)
{
  int i = 0;

  PrintString("****** Spanning Tree ******\n\n");

  for (i=0; i<nr_of_locs; i++) {
    PrintString("Location: ");
    PrintId(i+FIRST_LOCATION_ID);
    PrintString(":\n from_loc: ");
    PrintId(tree[i].from_loc);
    PrintString("\n from_dir: ");
    PrintId(tree[i].from_dir);
    PrintString("\n level: ");
    PrintNumber(tree[i].level);
    PrintString("\n\n");
  }
  PrintString("***************************\n\n");
}


void PrintRoute(int32_t *route)
{
  int nr_of_hops = 0;
  int i = 0;

  PrintString("****** Route ******\n\n");

  /* find starting location */
  while (route[nr_of_hops] != NO_ID)
    nr_of_hops++;

     for (i=nr_of_hops-1; i>=0; i--) {
       PrintId(route[i]);
       PrintString(", ");
       Output(outputbuffer);
     }
  PrintString("\n*******************\n\n");
  Output(outputbuffer);
}


void PrintUndoStack()
{
  int  i = undo_sp;
  char text_to_print[OUTPUT_LINE_LEN];

  PrintString("\n***** undo stack *****\n");
  if (undo_stack == NULL) {
    PrintString("   uninitialized undo stack\n********\n");
    return;
  }

  if (undo_sp == -1) {
    PrintString("   empty undo stack\n********\n");
    return;
  }

  sprintf(text_to_print, "position\titem1\titem2\titem3\titem4\tvalue\t(sp = %d)\n", undo_sp);
  PrintString(text_to_print);

  for (i=0; i<undo_stack_size; i++) {
    sprintf(text_to_print, "%d\t\t%d\t%d\t%d\t%d\t%d\n",
                           i,
                           undo_stack[i].item1,
                           undo_stack[i].item2,
                           undo_stack[i].item3,
                           undo_stack[i].item4,
                           undo_stack[i].value);
    PrintString(text_to_print);
  }
  return;
  /* undo_stack runs from 0 to undo_stack_size-1 */
  while (1) {
    sprintf(text_to_print, "%d\t\t%d\t%d\t%d\t%d\t%d\n",
                           i,
                           undo_stack[i].item1,
                           undo_stack[i].item2,
                           undo_stack[i].item3,
                           undo_stack[i].item4,
                           undo_stack[i].value);
    PrintString(text_to_print);
    if (i == undo_stack_size - 1) {
      /* roll over */
      i = 0;
    }
    else {
      i++;
    }

    if (i == undo_sp) {
      /* second time i equals undo_sp, means */
      /* that we printed the whole stack     */
      PrintString("**********************\n");
      return;
    }
  }
}


void PrintChoices(void)  /* @!@ */
{
  int  i = 0;
  char text_to_print[OUTPUT_LINE_LEN];

  PrintString("\n\n**** choices ****\n");
  PrintString("Following choices are available:\n\n");

  while (i<MAX_CHOICES) {
    if (choices[i].choice != NULL) {
      sprintf(text_to_print, "Choice, response %d: %s --- %s.\n\n", i, choices[i].choice, choices[i].response);
      PrintString(text_to_print);
    }
    else {
      PrintString("<empty choice>\n\n");
    }
    i++;
  }
  if (i==0) {
    PrintString("There are no choices.\n\n");
  }

  PrintString("*****************\n\n");
  Output(outputbuffer);
}

