
/************************************************************************/
/* Copyright (c) 2016 - 2022 Marnix van den Bos.                        */
/*                                                                      */
/* <marnix@xvan.nl>                                                     */
/*                                                                      */
/* This file is part of XVAN, an Interactive Fiction authoring system.  */
/*                                                                      */
/* XVAN is free software: you can redistribute it and/or modify         */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, version 3.                             */
/*                                                                      */
/* XVAN is distributed in the hope that it will be useful,              */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */
/*                                                                      */
/* You should have received a copy of the GNU General Public License    */
/* along with XVAN.  If not, see <http://www.gnu.org/licenses/>.        */
/*                                                                      */
/************************************************************************/


#include <stdio.h>
#include <stdlib.h>   /* malloc() */
#include <time.h>
#include <string.h>
#include <stdint.h>

#include "which_os.h"

#ifdef __windows_os
  #include "glk_win.h"
#else
  #include "glk_linux.h"
#endif

#include "keyword.h"
#include "typedefs.h"
#include "init.h"

/***********/
/* Globals */
/***********/

winid_t      mainwin   = NULL;        /* Glk story windows */
winid_t      statuswin = NULL;        /* Glk status window */

FILE         *datafile       = NULL;  /* storyfile             */
FILE         *transcriptfile = NULL;  /* to store transcript   */
FILE         *testfile       = NULL;  /* to read testinput     */

short        debug_info      = 0;
short        debug_level     = 0;

storyInfo    story_info;
int32_t      *stack;     /* Needed for executing triggers. */
int32_t      sp = 0;     /* stackpointer                   */

char         *outputbuffer = NULL;

choiceStruct choices[MAX_CHOICES];    /* for choice and hybrid mode */

/*************************/
/* function declarations */
/*************************/

int32_t CompilerVersionOK(void);
int32_t ExitProgram(void);

/************************/
/* Function definitions */
/************************/

int32_t CompilerVersionOK(void)
{
  if (strcmp(story_info.compiler_version, "2.6") != 0) {  /* @!@ */
    PrintError(37, NULL, NULL);
    PrintError(38, NULL, story_info.compiler_version);
    return(ERROR);
  }

  /* correct compiler version */
  return(OK);
}


int32_t ExitProgram(void)
{
  char ch;
  /* we need to exit, but don't close the console window immediately */

  /* flush outputbuffer */
  Output(outputbuffer);

  /* close transcript file */
  if (transcript) {
    transcript = 0;
    fclose(transcriptfile);
  }

  PrintError(39, NULL, NULL);
  GetGlkChar(&ch);

  /* Free all Mallocs */
  CleanUp();

  return(OK);
}


/************************/
/*         main()       */
/************************/

void glk_main(void)
{
  int i = 0;

  /* first create the windows */
  if (!CreateGlkWindows()) {
    /* creating windows failed, we cannot print an error msg */
    return;
  }

  /* init random number generator */
  srand((unsigned) time(NULL));

  /* malloc space for stack            */
  if ((stack = (int32_t *) malloc(STACK_SIZE*sizeof(int32_t))) == NULL) {
    PrintError(15, NULL, "glk_main()");
    ExitProgram();
    return;
  }

  /* init the choice array                 */
  /* must do that here, because GetInput() */
  /* will use ResetString() onm it.        */
  for (i=0; i<MAX_CHOICES;i++) {
    choices[i].choice   = NULL;
    choices[i].response = NULL;
  }

  /* Open the datafile */
  if ((datafile = fopen(DATAFILE, "rb")) == NULL) {
    PrintError(40, NULL, DATAFILE);
    ExitProgram();
    return;
  }

  if (!ReadStoryInfo(&story_info)) {
    /* error was printed by ReadStoryInfo() */
    ExitProgram();
    return;
  }

  /* First, check the compiler version in the story_info struct */
  if (!CompilerVersionOK()) {
    ExitProgram();
    return;
  }

  /* Read the directories from the datafile */
  if (!InitDirs()) {
    ExitProgram();
    return;
  }

  /* Read the max possible number of verb default codes.  */
  /* Can only do so AFTER the verb dir has been read.     */
  if (!InitVerbs()) {
    ExitProgram();
    return;
  }

  /* Read the max possible number of locations. */
  /* Can only do so AFTER the location dir has been read. */
  if (!InitLocations()) {
    ExitProgram();
    return;
  }

  /* Read the max possible number of objects. */
  /* Can only do so AFTER the object dir has been read. */
  if (!InitObjects()) {
    ExitProgram();
    return;
  }

  /* Read attributes.                   */
  /* Can only do this AFTER InitDirs(), */
  /* InitLocations() and InitObjects(). */
  if (!InitAttributes()) {
    ExitProgram();
    return;
  }

  if (!InitUndoStack()) {
    ExitProgram();
    return;
  }

  /* OK, we're all set. Let's roll! */
  LetsPlay();
  ExitProgram();
  return;
}
