
/************************************************************************/
/* Copyright (c) 2016 - 2022 Marnix van den Bos.                        */
/*                                                                      */
/* <marnix@xvan.nl>                                                     */
/*                                                                      */
/* This file is part of XVAN, an Interactive Fiction authoring system.  */
/*                                                                      */
/* XVAN is free software: you can redistribute it and/or modify         */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, version 3.                             */
/*                                                                      */
/* XVAN is distributed in the hope that it will be useful,              */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */
/*                                                                      */
/* You should have received a copy of the GNU General Public License    */
/* along with XVAN.  If not, see <http://www.gnu.org/licenses/>.        */
/*                                                                      */
/************************************************************************/


#include <stdio.h>
#include <string.h>
#include <stdint.h>

#include "which_os.h"
#ifdef __windows_os
  #include "glk_win.h"
#else
  #include "glk_linux.h"
#endif

#include "keyword.h"
#include "typedefs.h"
#include "input.h"

/*************************/
/* Function declarations */
/*************************/

void PrintPrompt(char[]);
char *GetInput(char[], char[]);
void GetGlkInput(char*);
void GetGlkChar(char*);

/************************/
/* Function definitions */
/************************/

void PrintPrompt(char prompt[])
{
  outputbuffer = ResetString(outputbuffer);
  PrintString("\n\n");
  PrintString(prompt);
  Output(outputbuffer);
  capital    = 1;

  return;
}


char *GetInput(char line_buf[], char prompt[])
{
  char text_to_print[OUTPUT_LINE_LEN];
  int  i = 0;

  /* Input is either read from the */
  /* commandline or from a file    */

  /* reset the choices array, as XeqAddChoice()     */
  /* wil check for NULL for next empty position     */
  /* Note: choices must be properly initilialized   */
  /* (NULL or valid string) otherwise resetString() */
  /* will crash                                     */
  for (i=0; i<MAX_CHOICES; i++) {
    choices[i].choice   = ResetString(choices[i].choice);
    choices[i].response = ResetString(choices[i].response);
  }

  /* flush the outputbuffer */
  Output(outputbuffer);

  if (testmode) {
    fgets(line_buf, INPUT_LINE_LEN, testfile);
      if (!feof(testfile)) {
      /* overwrite \n character with '\0' character */
      /* if we leave \n in place, looking up the    */
      /* word will fail.                            */

      PrintPrompt(prompt);
      line_buf[strlen(line_buf)-1] = '\0';
      sprintf(text_to_print, "%s\n", line_buf);
      PrintString(text_to_print);
      Output(outputbuffer);
      /* Output("\n"); */
    }
    else {
      /* we reached the end of the testfile*/
      PrintError(43, NULL, NULL); /* not an error */
      /* turn off testmode */
      testmode = 0;
      /* display the prompt */
      PrintPrompt(prompt);
    }
  }
  /* We do not use else here, because in case we read an */
  /* eof in testmode, we must ask the user for input to  */
  /* prevent an empty string to be returned, which will  */
  /* cause an 'I do not understand...' error.            */

  if (!testmode) {
    /* check if we must update the choices menu */
    /* for choice or hybrid mode                */
    if (story_info.play_mode != INTERPRETER_MODE) {
      /* UpdateChoices() also returns the choice in line_buf */
      UpdateChoicesMenu();

      /* get the choice using the story language */
      GetChoice(line_buf);
    }
    else {
      PrintPrompt(prompt);
      GetGlkInput(line_buf);
    }

    if (transcript) {
      WriteTranscript(line_buf);
      /* add <CR>. */
      /* we should check fprintf() again for errors, but we don't */
      /* the last one went ok, so assume this one will as well.   */
      WriteTranscript("\n");
    }
  }
  return(line_buf);
}

void GetGlkInput(char *line_buf)
{
  int     gotline;
  int     len;
  event_t ev;

  glk_request_line_event(mainwin, line_buf, 255, 0);
  gotline = 0;

  while (!gotline) {
    glk_select(&ev);
    switch (ev.type) {
      case evtype_LineInput:
        if (ev.win == mainwin) {
          len = ev.val1;
          line_buf[len] = '\0';
          gotline = 1;
        }
      break;
      case evtype_Arrange:
        /* Do nothing */
      break;
    }
  }
}


void GetGlkChar(char *ch)
{
  int gotchar;
  event_t ev;

  glk_request_char_event(mainwin);
  gotchar = 0;
  while (!gotchar) {
    glk_select(&ev);
    switch (ev.type) {
      case evtype_CharInput:
        if (ev.win == mainwin) {
          *ch = ev.val1;
          gotchar = 1;
        }
      break;
      case evtype_Arrange:
        /* do nothing */
      break;
    }
  }
}
