
/************************************************************************/
/* Copyright (c) 2016, 2017, 2018, 2019 Marnix van den Bos.             */
/*                                                                      */
/* <marnix.home@gmail.com>                                              */
/*                                                                      */
/* This file is part of XVAN, an Interactive Fiction authoring system.  */
/*                                                                      */
/* XVAN is free software: you can redistribute it and/or modify         */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, version 3.                             */
/*                                                                      */
/* XVAN is distributed in the hope that it will be useful,              */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */
/*                                                                      */
/* You should have received a copy of the GNU General Public License    */
/* along with XVAN.  If not, see <http://www.gnu.org/licenses/>.        */
/*                                                                      */
/************************************************************************/

#include <stdio.h>
#include <stdlib.h>   /* free()    */
#include <string.h>   /* strlen()  */
#include <ctype.h>    /* tolower() */
#include <stdint.h>

/* DO NOT INCLUDE keyword.h */

/*************************/
/* Function declarations */
/*************************/

int32_t SkipString(char*, int*);
int32_t ValidateJsonNumber(char*, int*);
int32_t ValidateJsonArray(char*, int*);
int32_t ValidateJsonObject(char*, int*);

/************************/
/* Function definitions */
/************************/

int32_t SkipString(char *string, int *index)
{
  /* string[*index] is start quote */
  do {
    (*index)++;
  } while (string[*index] != '"' && string[*index] != '\0');

  return(string[*index] == '"' ? 1 : 0);
}


int32_t ValidateJsonNumber(char *json_number, int *index)
{
  int state = 1;

  /* this function checks whether number_string */
  /* complies with the json number syntax       */

  while (1) {
    switch (json_number[*index]) {
      case '-':
        switch (state) {
          case 0:
            state = 1;
            break;
          case 5:
            state = 6;
            break;
          default:
            return(0);
        }
        break;
      case '+':
        switch (state) {
          case 5:
            state = 6;
            break;
          default:
            return(0);
        }
        break;
      case '.': ;
      case ',':
        switch (state) {
          case 2:
            state = 4;
            break;
          case 3:
            state = 4;
            break;
          default:
            return(0);
        }
        break;
      case 'e': ;
      case 'E':
        switch (state) {
          case 2:
            state = 5;
            break;
          case 3:
            state = 5;
            break;
          default:
            return(0);
        }
        break;
      case '0':
        switch (state) {
          case 0: ;
          case 1:
            state = 2;
            break;
          case 3: ;
          case 4: ;
          case 6:
            /* state remains the same */
            break;
          default:
            return(0);
        }
        break;
      case '1': ;
      case '2': ;
      case '3': ;
      case '4': ;
      case '5': ;
      case '6': ;
      case '7': ;
      case '8': ;
      case '9':
        switch (state) {
          case 3: ;
          case 4: ;
          case 6:
            /* state remains the same */
            break;
          default:
            return(0);
        }
        break;
      default:
        switch (state) {
          case 2: ;
          case 3: ;
          case 4: ;
          case 6:
            return(1);
          default:
            return(0);
        }
        break;
    } /* switch(number_string[*index]) */
    (*index)++;
  } /* while */
}


int32_t ValidateJsonArray(char *json_array, int *index)
{
  int i     = 0;
  int state = 1;

  /* this function checks whether json_string */
  /* complies with the json array syntax      */

  /* convert everything to lowercase */
  for (i=0; i<strlen(json_array); i++)
    json_array[i] = tolower(json_array[i]);

  /* we use this function recursive, so skip */
  /* the '[' to prevent an endless loop      */

  if (json_array[(*index)++] != '[')
    return(0);

  while (1) {
    switch (json_array[*index]) {
      case '\0':
        /* end of string before ']'*/
        return(0);
      case '{':
        switch (state) {
          case 1: ;
            state = 2;
            /* an object */
            if (!ValidateJsonObject(json_array+(*index), index))
              return(0);
            break;
          default:
            return(0);
        }
        break;
      case '[':
        switch (state) {
          case 1: ;
          case 3:
            state = 2;
            /* a nested array */
            if (!ValidateJsonArray(json_array+(*index), index))
              return(0);
            break;
          default:
            return(0);
        }
        break;
      case ' ':
        /* all states allow 1 or more spaces */
        break;
      case '"':
        switch (state) {
          case 1: ;
          case 3: ;
            state = 2;
            /* a string */
            if (!SkipString(json_array+(*index), index))
              return(0);
            break;
          default:
            return(0);
        }
        break;
      case 't':
        switch (state) {
          case 1: ;
          case 3:
            state = 2;
            /* a true */
            if (strncmp(json_array+(*index), "true", 4) != 0)
              return(0);
            break;
          default:
            return(0);
        }
        break;
      case 'f':
        switch (state) {
          case 1: ;
          case 3:
            state = 2;
            /* a false */
            if (strncmp(json_array+(*index), "false", 5) != 0)
              return(0);
            break;
          default:
            return(0);
        }
        break;
      case 'n':
        switch (state) {
          case 1: ;
          case 3:
            state = 2;
            /* a null */
            if (strncmp(json_array+(*index), "null", 4) != 0)
              return(0);
            break;
          default:
            return(0);
        }
        break;
      case ',':
        switch (state) {
          case 2:
            /* a comma */
            state = 3;
            break;
          default:
            return(0);
        }
        break;
      case ']':
        switch (state) {
          case 1: ;
          case 2:
            /* end of the array */
            return(1);
            break;
          default:
            return(0);
        }
        break;
      default:
        return(0);
    } /* switch(number_string[*index]) */
    (*index)++;
  } /* while */
}


int32_t ValidateJsonObject(char *json_string, int *index)
{
  int i     = 0;
  int state = 1;

  /* this function checks whether json_string */
  /* complies with the json object syntax     */

  /* convert everything to lowercase */
  for (i=0; i<strlen(json_string); i++)
    json_string[i] = tolower(json_string[i]);

  /* we use this function recursive, so skip */
  /* the '{' to prevent an endless loop      */

  if (json_string[(*index)++] != '{')
    return(0);

  while (1) {
    switch (json_string[*index]) {
      case '\0':
        /* end of string before '}'*/
        return(0);
      case '{':
        switch (state) {
          case 3:
            state = 4;
            /* a nested object */
            if (!ValidateJsonObject(json_string+(*index), index))
              return(0);
            break;
          default:
            return(0);
        }
        break;
      case ' ':
        /* all states allow 1 or more spaces */
        break;
      case '[':
        switch (state) {
          case 3:
            state = 4;
            /* an array */
            if (!ValidateJsonArray(json_string+(*index), index))
              return(0);
            break;
          default:
            return(0);
        }
        break;
      case '}':
        switch (state) {
          case 1: ;
          case 4:
            return(1);
          default:
            return(0);
        }
        break;
      case '"':
        switch (state) {
          case 1:
            state = 2;
            break;
          case 3:
            state = 4;
            break;
          case 5:
            state = 2;
            break;
          default:
            return(0);
        }
        /* a string */
        if (!SkipString(json_string+(*index), index))
          return(0);
        break;
      case 't':
        switch (state) {
          case 3:
            state = 4;
            /* a true */
            if (strncmp(json_string+(*index), "true", 4) != 0)
              return(0);
            break;
          default:
            return(0);
        }
        break;
      case 'f':
        switch (state) {
          case 3:
            state = 4;
            /* a false */
            if (strncmp(json_string+(*index), "false", 5) != 0)
              return(0);
            break;
          default:
            return(0);
        }
        break;
      case 'n':
        switch (state) {
          case 3:
            state = 4;
            /* a null */
            if (strncmp(json_string+(*index), "null", 4) != 0)
              return(0);
            break;
          default:
            return(0);
        }
        break;
      case ',':
        switch (state) {
          case 4:
            /* a comma */
            state = 5;
            break;
          default:
            return(0);
        }
        break;
      case ':':
        switch (state) {
          case 2:
            /* a colon */
            state = 3;
            break;
          default:
            return(0);
        }
        break;
      default:
        switch (state) {
          case 3:
            state = 4;
            /* check for a number */
            if (!ValidateJsonNumber(json_string+(*index), index))
              return(0);
            break;
          default:
            return(0);
        }
        break;
    } /* switch(json_string[*index]) */
    (*index)++;
  } /* while */
}
