
/************************************************************************/
/* Copyright (c) 2016 - 2022 Marnix van den Bos.                        */
/*                                                                      */
/* <marnix@xvan.nl>                                                     */
/*                                                                      */
/* This file is part of XVAN, an Interactive Fiction authoring system.  */
/*                                                                      */
/* XVAN is free software: you can redistribute it and/or modify         */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, version 3.                             */
/*                                                                      */
/* XVAN is distributed in the hope that it will be useful,              */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */
/*                                                                      */
/* You should have received a copy of the GNU General Public License    */
/* along with XVAN.  If not, see <http://www.gnu.org/licenses/>.        */
/*                                                                      */
/************************************************************************/


#include <stdio.h>
#include <stdlib.h>   /* malloc() */
#include <string.h>
#include <stdint.h>

#include "keyword.h"
#include "typedefs.h"
#include "letsplay.h"

/***********/
/* Globals */
/***********/

/* these must be defined as externals in the header files of the */
/* modules that need them.                                       */

short testmode    = 0;            /* read input from file is off  */
short transcript  = 0;            /* copy output to file is off   */

char  prompt[MAX_PROMPT_LEN+1];

int32_t active_entity = NO_ID;    /* Location or object that is   */
                                  /* currently active (especially */
                                  /* for use with common          */
                                  /* triggers as 'this').         */
int32_t curr_loc      = NO_ID;    /* Current location.            */
int32_t direction     = NO_ID;    /* Current direction.           */
int32_t prepos        = NO_ID;    /* Last preposition.            */
int32_t action        = NO_ID;    /* Current action.              */
int32_t actor         = PLAYER;   /* Current actor.               */
int32_t subject       = NO_ID;    /* Current subject(s).          */
int32_t specifier     = NO_ID;    /* Current specifier.           */
int32_t value         = 0;        /* Current value.               */
int32_t ordinal       = 0;        /* Current ordinal value.       */

int32_t indent        = 0;        /* For indentation.                          */
int32_t capital       = 1;        /* Start printing with a capital character.  */
int32_t article       = 0;        /* Tells whether to print an article.        */
int32_t muted         = 0;        /* Tells whether to print the output         */

/*************************/
/* Function declarations */
/*************************/

int32_t LetsPlay(void);

/************************/
/* Function definitions */
/************************/

int32_t LetsPlay(void)
{
  char         line_buf[INPUT_LINE_LEN];
  parsedInput  parsed_input;
  usrActionRec action_rec;
  int32_t      parse_syntax  = LINE; /* Tells what syntax to parse. */
  int32_t      subject_index = 0;    /* Tells which subject to use  */
                                     /* from action_rec.            */
  int32_t      cont_list[nr_of_locs+nr_of_objs];

  int32_t      parse_result; /* Needed to process result of ParseInput(). */
  resultStruct translate_result = {OK, NONE, 0};
  resultStruct arec_result;
  resultStruct prologue_result;
  resultStruct verb_def_result;
  resultStruct unbound_result;   /* for unbound subjects */  /* @!@ */
  attrInfo     *attributes = NULL;  /* @!@ */
  int          index = 0;  /* @!@ */
  int32_t      i = 0;

  /* 17dec2016: added nr_of_subjects because introducing the parser  */
  /* rules makes it no longer possible to match all subjects at once */
  /* in trnslate.c. The parser rules may need to relate subject to   */
  /* specifier for disambiguation, so each subject must bne matched  */
  /* individually now under control of letsplay().                   */

  int32_t nr_of_subjects = 0;

  /* set the prompt */
  if (debug_info)
    strncpy(prompt, "debug> ", MAX_PROMPT_LEN);
  else
    strncpy(prompt, "> ", MAX_PROMPT_LEN);

  /* Print interpreter opening message. */
  for (i=0; i<25; i++)
    printf("\n");

  outputbuffer = ResetString(outputbuffer);
  PrintString("-- XVAN 2.6 interpreter --\n\n");  /* @!@ */
  PrintString("(c) Marnix van den Bos\n\n\n\n");

  PrintString(story_info.title);

  switch(story_info.story_language) {
    case ENG:
      PrintString(" version ");
      PrintString(story_info.version);
      PrintString(" - English\n\n\n\n");
      break;
    case NL:
      PrintString(" versie ");
      PrintString(story_info.version);
      PrintString(" - Nederlands\n\n\n\n");
      break;
    default:
      /* we should never get here */
      PrintString(" version ");
      PrintString(story_info.version);
      PrintString(" - Unknown language\n\n\n\n");
      break;
  }

  /* now print the intro screen */
  Output(outputbuffer);

  /* Init current location. */
  curr_loc = obj_dir[PLAYER-FIRST_OBJECT_ID].held_by;

  /* Init line_buf. */
  line_buf[0] = '\0';

  /* Handle the timers to allow starting of the game   */
  /* (printing opening messages etc).                  */
  /* 06oct2017 added action_rec and subject_index, but */
  /* in this first call of HandleTimers() they have    */
  /* no valid values yet                               */
  /* Make sure no undo info is written                 */
  write_undo = 0;
  if (HandleTimers(&action_rec, subject_index) == QUIT)
    return(OK);

  while (1) {
    /* check if an end of undo for the previous turn must be written */
    if (write_undo) {
      PushUndoItem(EOU, NO_ID, NO_ID, NO_ID, NO_ID, 0);
    }

    /* start each new turn enabling write undo information */
    write_undo = 1;

    /* start the turn with 0 undo records */
    undo_record_counter = 0;

    /* flush outputbuffer */
    Output(outputbuffer);

    /* Reset subject_index. */
    subject_index = 0;

    if (parse_syntax == LINE) {
      /* Fill the parsed input record with default values. */
      InitParsedInput(&parsed_input);

      /* Fill the action record with default values.       */
      InitUsrActionRec(&action_rec);
    }
    /* output must start with a capital letter */
    capital = 1;
/*UpdateChoicesMenu(line_buf);*/

    if (strlen(line_buf) == 0) {
      /* Read the user's input and skip comments */
      do {
        GetInput(line_buf, prompt);
      } while (line_buf[0] == PLAYER_COMMENT);
    }

    /* Parse the input. */
    parse_result = ParseInput(line_buf, &parsed_input, parse_syntax);

    /* possible return values: OK, ERROR, UNKNOWN_WORD */
    if (parse_result == ERROR && parse_syntax != LINE) {
      /* The user was asked for a subject or specifier, but */
      /* parsing his response failed. He may have entered a */
      /* next line. Parse again, but now with LINE syntax.  */
      /* If parse error, skip rest of while loop.           */

      /* Fill the parsed input record with default values. */
      InitParsedInput(&parsed_input);

      /* Fill the action record with default values.        */
      InitUsrActionRec(&action_rec);

      parse_result = ParseInput(line_buf, &parsed_input, LINE);
    }
    /* Reset parse_syntax. */
    parse_syntax = LINE;

    if (parse_result == OK) {

      /* Clear line_buf. */
      line_buf[0] = '\0';

      /* 17dec2016: we must translate for each subject            */
      /* individually because we have introduced the parser rules */
      /* where a subject may be checked against the specifier     */

      /* nr_of_subjects is the numer of subjects that were in the */
      /* command line. Subjects that result from a plural subject */
      /* are counted separately in the OK: branch from Translate() */

      nr_of_subjects = CountSubjectsInParsedInput(&parsed_input);

      /* if more than 1 subject we have to go through them       */
      /* backwards, because of the exteded system description    */
      /* e.g. give rope, bone to fred won't work if we don't     */
      /* start with 'bone to fred' as the first subject          */

      subject_index = (nr_of_subjects == 0 ? 0 : nr_of_subjects-1);

      /* 17dec2016: added iteration through subjects (in case more    */
      /* than 1 subject was entered by the player). Iteration used to */
      /* be done by Translate(), but with the introduction of parse   */
      /* rules the iteration had to be taken out of Translate()       */
      /* Translate() because of possible parse dependencies between   */
      /* subject and specifier.                                       */

      /* 03oct2019: now that we support plurality, there may be       */

      /* number of subjects may be zero, but we must go through */
      /* the loop at least once (e.g. "look", "north", ...)     */

      do {
        translate_result = Translate(&parsed_input, subject_index, &action_rec, line_buf);
        /* action_rec may contain the 'it' object. We must replace */
        /* it by the actual value before we can start comparing    */
        if (!ReplaceItObjects(&action_rec))
          return(ERROR);

        switch (translate_result.tag) {
          case OK:
            /* Successful translation.                             */

            /* if the subject was plural, Translate() may have     */
            /* returned more than 1 matching subject id. If this   */
            /* is the case, we must execute this "case OK:" clause */
            /* for all returned ids.                               */

            /* Cannot use 'while' here, because we must go through */
            /* the loop if the first subject equals NO_ID. If one  */
            /* of the other subjects equals NO_ID, we must stop.   */

            i = 0;

            do {
              /* Set the global subject variable. */
              if (action_rec.subject[subject_index] == PLURAL) {
                subject = action_rec.plural_subjects[i++];
              }
              else {
                subject = action_rec.subject[subject_index];
              }

              /* Now execute the verb's prologue, if any. */

              prologue_result = XeqPrologue(action_rec.action1);
              switch (prologue_result.tag) {
                case QUIT:
                  /* exit */
                  return(OK);
                case DISAGREE:
                  /* MUST WE EXECUTE THE EPILOGUE HERE ??? */
                  /* Forget the whole thing, continue with */
                  /* next user input.                      */

                  /* Check if we must handle the timers. */
                  if (CheckDoTimers()) {
                    if (HandleTimers(&action_rec, subject_index) == QUIT)
                      return(OK);
                  }
                  break;
                case GET_SUBJECT:
                  /* The subject wasn't specified in the input.  */
                  /* prologue_result.value may contain a word id */
                  /* that must be copied to the prepositions in  */
                  /* the parsed_input struct                     */
                  if (prologue_result.value != NO_ID) {
                     if (parsed_input.prepositions.nr_of_prepositions != MAX_PARSE_PREPOS) {
                      /* room for an extra preposition*/
                      parsed_input.prepositions.preposition[parsed_input.
                           prepositions.nr_of_prepositions++] = prologue_result.value;
                    }
                  }
                  parse_syntax = SUBJECT;
                  break;
                case GET_SPECIFIER:
                  /* The specifier wasn't specified in the input. */
                  if (prologue_result.value != NO_ID) {
                     if (parsed_input.prepositions.nr_of_prepositions != MAX_PARSE_PREPOS) {
                      /* room for an extra preposition*/
                      parsed_input.prepositions.preposition[parsed_input.
                           prepositions.nr_of_prepositions++] = prologue_result.value;
                    }
                  }
                  parse_syntax = SPECIFIER;
                  break;
                case GET_ANSWER:
                  /* We need an answer from the player. */
                  parse_syntax = ANSWER;
                  break;
                default:
                  /* Either AGREE or NO_MATCH. */
                  /* Build list with contained objs for curr_loc. */
                  if (!ContList(curr_loc, cont_list, parsed_input.scope))
                    return(OK); /* actually, is an error */

                  /* Execute action_rec for list. */
                  arec_result = XeqActionRec(&action_rec, cont_list, subject_index);
                  switch (arec_result.tag) {
                    case NO_MATCH:
                      /* No match; execute the default verb code. */
                      verb_def_result = XeqVerbDefault(&action_rec, subject_index);
                      switch (verb_def_result.tag) {
                        case QUIT:
                          /* Stop */
                          return(OK);
                        case GET_SUBJECT:
                          /* The subject wasn't specified in the input. */
                          /* verb_def_result.value may contain a word id  */
                          /* that must be copied to the prepositions in   */
                          /* the parsed_input struct                      */
                          if (verb_def_result.value != NO_ID) {
                            if (parsed_input.prepositions.nr_of_prepositions != MAX_PARSE_PREPOS) {
                              /* room for an extra preposition*/
                              parsed_input.prepositions.preposition[parsed_input.
                                   prepositions.nr_of_prepositions++] = verb_def_result.value;
                            }
                          }
                          parse_syntax = SUBJECT;
                          break;
                        case GET_SPECIFIER:
                          /* The specifier wasn't specified in the input. */
                          /* verb_def_result.value may contain a word id  */
                          /* that must be copied to the prepositions in   */
                          /* the parsed_input struct                      */
                          if (verb_def_result.value != NO_ID) {
                            if (parsed_input.prepositions.nr_of_prepositions != MAX_PARSE_PREPOS) {
                              /* room for an extra preposition*/
                              parsed_input.prepositions.preposition[parsed_input.
                                   prepositions.nr_of_prepositions++] = verb_def_result.value;
                            }
                          }
                          parse_syntax = SPECIFIER;
                          break;
                        case GET_ANSWER:
                          /* We need an answer from the player. */
                          parse_syntax = ANSWER;
                          break;
                        default:
                          /* Either AGREE, DISAGREE or NO_MATCH.   */
                          /* 15jun2019 - added NO_MATCH to default */
                          /* it first had its own branch, but did  */
                          /* not execute the epilogue.             */
                          XeqEpilogue(action_rec.action1);

                         /* Check if we must handle the timers. */
                          if (CheckDoTimers()) {
                            if (HandleTimers(&action_rec, subject_index) == QUIT)
                              return(OK);
                          }
                          break;
                      } /* switch XeqVerbDefault() */
                      break;
                    case QUIT:
                      /* Stop */
                      return(OK);
                    case GET_SUBJECT:
                      /* The subject wasn't specified in the input. */
                      /* arec_result.value may contain a word id that */
                      /* must be copied to the prepositions in the    */
                      /* parsed_input struct                          */
                      if (arec_result.value != NO_ID) {
                        if (parsed_input.prepositions.nr_of_prepositions != MAX_PARSE_PREPOS) {
                          /* room for an extra preposition*/
                          parsed_input.prepositions.preposition[parsed_input.
                               prepositions.nr_of_prepositions++] = arec_result.value;
                        }
                      }
                      parse_syntax = SUBJECT;
                      break;
                    case GET_SPECIFIER:
                      /* The specifier wasn't specified in the input. */
                      /* arec_result.value may contain a word id that */
                      /* must be copied to the prepositions in the    */
                      /* parsed_input struct                          */
                      if (arec_result.value != NO_ID) {
                        if (parsed_input.prepositions.nr_of_prepositions != MAX_PARSE_PREPOS) {
                          /* room for an extra preposition*/
                          parsed_input.prepositions.preposition[parsed_input.
                               prepositions.nr_of_prepositions++] = arec_result.value;
                        }
                      }
                      parse_syntax = SPECIFIER;
                      break;
                    case GET_ANSWER:
                      /* We need an answer from the player. */
                      parse_syntax = ANSWER;
                      break;
                    default:
                      /* AGREE or DISAGREE     */
                      /* Execute the epilogue. */
                      XeqEpilogue(action_rec.action1);

                      /* Check if we must handle the timers. */
                      if (CheckDoTimers()) {
                        if (HandleTimers(&action_rec, subject_index) == QUIT)
                          return(OK);
                      }

                      parse_syntax = LINE;
                      break;
                  } /* switch XeqActionRec() */
                  break;
              } /* switch XeqPrologue() */
            } while (action_rec.subject[subject_index] == PLURAL &&
                     action_rec.plural_subjects[i] != NO_ID);
            break; /* Translate() OK */
          case OVERFLOW:
            /* More than MAX_HITS hits in the Find() function that */
            /* is called by Translate()                            */
            /* too many hits, use adjectives to narrow down */
            PrintError(45, NULL, NULL);
            break;
          case NEXT_SENTENCE:
            /* next sentence was entered, do nothing */
            break;
          case UNKNOWN_WORD:
            line_buf[0] = '\0';
            break;
          case ERROR:
            /* error msg already printed */
            break;
          case ACTOR_ERROR:
            PrintNotFound(&(parsed_input.actor));
            break;
          case SPECIFIER_ERROR:
            PrintNotFound(&(parsed_input.specifier));
            break;
          case SUBJECT_ERROR:  /* @!@ */
            /* check for an unbound subject. they may have entered a  */
            /* a sentence like "get some rest" or "hit the road". In  */
            /* such a case there will be no matching object.          */
            /* Conditions to process unbound subject:                 */
            /*   - the user input may only have 1 subject             */
            /*   - the subject must be a single noun, no adjectives   */
            /*   - still to decide whether a specifier may be present */
            /*      => o_unbound checks for empty specifier           */

            if (nr_of_subjects == 1 && parsed_input.subject[0].part1.nr_of_adjectives == 0) {
              /* check for unbound subject                              */
              /* load the unbound noun in attribute o_unbound.r_unbound */
              if (!GetAttributeInfo(UNBOUND_ATTR, UNBOUND, &attributes, &index)) {
                /* do nothing */
                break;
              }

              /* no store the noun value in the attribute */
              attributes[index].type  = WORD_ID;
              attributes[index].value = parsed_input.subject[0].part1.noun;

              /* now call trigger o_unbound.t_unbound */
              unbound_result = XeqTrigger(UNBOUND, T_UNBOUND, &action_rec, 0);
              /* must return AGREE or DISAGREE, no_match is not an option */
              if (unbound_result.tag == AGREE) {
                break;
              }
            }

            /* no unbound subject, print "You cannot see ..." message */
            PrintNotFound(&(parsed_input.subject[translate_result.value]));
            break;
        } /* switch translate_result */
        subject_index--;
      } while (subject_index >= 0);
    } /* if (parse_result == OK)*/
    else {
      /* Parse error, reset line_buf. */
      /* Error msg (either unknown word or syntax error) has */
      /* already been printed .                              */
      if (parse_result != UNKNOWN_WORD) {
        /* unknown word error has already been printed by    */
        /* CheckSyntax().                                    */
        /* I don't understand that sentence */
        PrintError(46, NULL, NULL);
      }
      line_buf[0] = '\0';
    }
  } /* while (1) */
  /* No return here, make compiler happy. */
}
