
/************************************************************************/
/* Copyright (c) 2016 - 2022 Marnix van den Bos.                        */
/*                                                                      */
/* <marnix@xvan.nl>                                                     */
/*                                                                      */
/* This file is part of XVAN, an Interactive Fiction authoring system.  */
/*                                                                      */
/* XVAN is free software: you can redistribute it and/or modify         */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, version 3.                             */
/*                                                                      */
/* XVAN is distributed in the hope that it will be useful,              */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */
/*                                                                      */
/* You should have received a copy of the GNU General Public License    */
/* along with XVAN.  If not, see <http://www.gnu.org/licenses/>.        */
/*                                                                      */
/************************************************************************/


#include <stdio.h>
#include <stdlib.h>   /* malloc()  */
#include <string.h>
#include <ctype.h>    /* toupper() */
#include <stdint.h>

#include "which_os.h"
#ifdef __windows_os
  #include "glk_win.h"
#else
  #include "glk_linux.h"
#endif

#include "keyword.h"
#include "typedefs.h"
#include "output.h"

#define CR (char) 10

/* The non-Glk version of the interpreter has a mechanism that   */
/* ensures that words that don't fit on the remainder of the     */
/* current line, get printed on the next line and do not end up  */
/* broken in half (using the char_count variable).               */
/* The Glk version of the interpreter does't need such a         */
/* mechanism, as word wrapping is controlled by hte Glk library. */
/* The non-Glk interpreter's mechanism could be used  for        */
/* grid-windows if we want to control word wrapping in a grid.   */

/*************************/
/* Function declarations */
/*************************/

int      CreateGlkWindows(void);
void     SetGlkStyles(void);
char     *ResetString(char*);
char     *ExpandString(char*, int);
char     *AddToString(char*, char*);
void     WriteTranscript(char*);
void     Output(char*);
void     PrintString(char*);
void     PrintWord(int32_t);
void     PrintExtendedSysDescr(extendedSysDescr*);
void     PrintSysDescr(sysDescr*);
void     PrintArticle(sysDescr*, int32_t);
void     PrintId(int32_t);
void     PrintNumber(int32_t);
int32_t  IsSpecialChar(char);
int32_t  IsVowel(char);
int32_t  Power(int32_t, int32_t);

/************************/
/* Function definitions */
/************************/

/* 03Feb2018: with function try() we allow to */
/* mute output, so nothing will be printed.   */
/* When mute is on, we must not change the    */
/* char_count var because we don't print      */
/* Therefore, each function in this module    */
/* that manipulates char_count, has an early  */
/* exit when muted == 1                       */


void SetGlkStyles(void)
{
  /* set styles text color, background color and bold font size for styles Normal, Alert and Note */
  glk_stylehint_set(wintype_TextBuffer, style_Normal, stylehint_TextColor, 0x00FFFFFF);
  glk_stylehint_set(wintype_TextBuffer, style_Normal, stylehint_BackColor, 0x00000084);

  glk_stylehint_set(wintype_TextBuffer, style_Alert, stylehint_TextColor, 0x00FFFFFF);
  glk_stylehint_set(wintype_TextBuffer, style_Alert, stylehint_BackColor, 0x00000084);

  glk_stylehint_set(wintype_TextBuffer, style_Note, stylehint_TextColor, 0x00FFFFFF);
  glk_stylehint_set(wintype_TextBuffer, style_Note, stylehint_BackColor, 0x00000084);

  glk_stylehint_set(wintype_TextBuffer, style_Alert, stylehint_Size, -1);

  glk_stylehint_set(wintype_TextGrid, style_Normal, stylehint_TextColor, 0x00000084);
  glk_stylehint_set(wintype_TextGrid, style_Normal, stylehint_BackColor, 0x00FFFFFF);

  glk_stylehint_set(wintype_TextGrid, style_Alert, stylehint_TextColor, 0x00000084);
  glk_stylehint_set(wintype_TextGrid, style_Alert, stylehint_BackColor, 0x00FFFFFF);

  glk_stylehint_set(wintype_TextGrid, style_Note, stylehint_TextColor, 0x00000084);
  glk_stylehint_set(wintype_TextGrid, style_Note, stylehint_BackColor, 0x00FFFFFF);

  glk_stylehint_set(wintype_TextGrid, style_Normal, stylehint_Size, -1);
  glk_stylehint_set(wintype_TextGrid, style_Note, stylehint_Size, -1);
  glk_stylehint_set(wintype_TextGrid, style_Alert, stylehint_Size, -2);

  glk_stylehint_set(wintype_TextBuffer, style_Input, stylehint_Weight, 0);

  /* stylehints for user input representation */
  glk_stylehint_set(wintype_TextBuffer, style_Input, stylehint_TextColor, 0x00FFFFFF);
  glk_stylehint_set(wintype_TextBuffer, style_Input, stylehint_BackColor, 0x00000084);
}


int CreateGlkWindows(void)
{
   /* set styles text color, background color and bold font size for styles Normal, Alert and Note */
  SetGlkStyles();

  /* Open the main window. */
  mainwin = glk_window_open(0, 0, 0, wintype_TextBuffer, 1);
  if (!mainwin) {
    /* failed to open. */
    return(ERROR);
  }

  /* Set the current output stream to print to it. */
  glk_set_window(mainwin);

  /* Open a second window: a text grid, above the main window, three lines
      high. It is possible that this will fail also, but we accept that. */
  statuswin = glk_window_open(mainwin, winmethod_Above | winmethod_Fixed,
      3, wintype_TextGrid, 0);
  if (!statuswin) {
    return(ERROR);
  }

  return(OK);
}


char *ResetString(char *string)
 /* string must be maloc-ed */
{
  if (string == NULL) {
    return(NULL);
  }
  else {
    free(string);
    return(NULL);
  }
}


char *ExpandString(char *source, int amount)
{
  int  len = 0;
  char *temp;

  if (!source) {
    /* amount + 1 because of the '\0' */
    if ( (source = (char*) malloc((amount+1)*sizeof(char))) == NULL )  {
      return(NULL);
    }
    source[0] = '\0';
  }
  else {

    /* don't know why the F#$K realloc() crashes the debugger        */
    /* got the infamous "cannot find the bounds of current function" */

    /* my own realloc() */
    len = strlen(source);
    if ( !(temp = (char *) malloc((len+1)*sizeof(char))) )
      return(NULL);
    strcpy(temp, source);
    free(source);
    if ( !(source = (char *) malloc((len+amount+1)*sizeof(char))) )  /* +1 for the '\0\ */
      return(NULL);
    strcpy(source, temp);
    free(temp);
  }
  return(source);
}


char *AddToString(char *dest, char *to_add)
{
  /* dest must be malloced */

  int len = 0;

  /* 21may2018: design decision: in case   */
  /* dest is NULL, we will first create an */
  /* empty string.                         */
  if (dest == NULL) {
    dest = ExpandString(dest, 0);
  }

  if (to_add == NULL) {
    return(dest);
  }

  if ( (len = strlen(to_add)) == 0 ) {
    return(dest);
  }

  /* ExpandString() will take care of '\0' in dest */
  if ( !(dest = ExpandString(dest, len))) {
    return(NULL);
  }

  strncat(dest, to_add, len);

  return(dest);
}


void WriteTranscript(char *text)
{
  if (transcript) {
    if (fprintf(transcriptfile, text) < 0) {
      PrintError(44, NULL, NULL);
      /* not a severe error, continue but turn off transcript mode */
      transcript = 0;
    }
    /* add <CR>. */
    /* we should check fprintf() again for errors, but we don't */
    /* the last one went ok, so assume this one will as well.   */
    fprintf(transcriptfile, "\n");
  }
  else {
    /* transcript mode is not turned on */
    PrintError(107, NULL, NULL);
  }
}


void Output(char *line)
 /* line must be malloc-ed */
{
  /* glk_put_string() does not like NULL */

  if (line == NULL) {
    return;
  }

  if (muted == 1) {
    outputbuffer = ResetString(line);
    return;
  }

  glk_put_string(line);

  if (transcript) {
    WriteTranscript(line);
  }

  outputbuffer = ResetString(line);
}


void PrintString(char *str)
{
  /* Syntax: <string> or                            */
  /*         [string]<-1><parameter>[string]        */

  /* Printstring takes into account the output line */
  /* length (OUTPUT_LINE_LEN) and makes sure that   */
  /* a word does not get cut in half thru wrapping  */

  int32_t i    = 0;
  int32_t id1  = NO_ID;                /* Needed in case str contains an id. */
  int32_t id2  = NO_ID;
  int32_t type = NO_TYPE;              /* dummy for ConvSpecId()             */
  char word_buffer[OUTPUT_LINE_LEN];   /* To build up a full word before     */
                                       /* printing it                        */
  char text_to_print[OUTPUT_LINE_LEN]; /* not sure if we need this           */
  int32_t j = 0;                       /* index for word_buffer              */
  int32_t cont = 1;
  int32_t owner = NO_ID;
  char    *descr = NULL;               /* to retrieve a description          */

  attrInfo *attributes = NULL;         /* only need the pointer */
  int32_t  attribute_index = 0;

  if (muted == 1)
    return;

  word_buffer[0] = '\0';

  /* Concatenate single chars to a word and then check if it still fits on  */
  /* the current line. If not, start at the next line. ALWAYS stop at a -1. */
  while (cont) {
    if (str[i] == '\0')
      cont = 0;
    if (str[i] != (char)-1) {  /* &&& */ /* else added  and removed again */
      word_buffer[j++] = (capital == 1 ? (char) toupper((int32_t) str[i]) : str[i]);

      /* check to see if capital variable must be set or cleared */
      if (str[i] == '.' || str[i] == '?' || str[i] == '!')
        /* after a '.', '?' or '!' a new sentence starts */
        capital = 1;
      else if (str[i] != ' ' && str[i] != '\0' && str[i] != CR)
        /* a space or CR may come after a dot; to be on the safe side, */
        /* never change the value of capital after a white space, a CR */
        /* or a NULL character.                                        */
        capital = 0;

      if (IsSpecialChar(str[i])) {
        /* if we have a white space and a pending article, the white space */
        /* must not be processed to prevent double spaces when the object  */
        /* or location does not have an article associated with it.        */
        if (str[i] == SPACE && article != 0) {
          /* skip the space that was already added to str[] */
          j--;
        }

        /* check if the whole word fits on the remainder of the line       */
        word_buffer[j] = '\0';
        sprintf(text_to_print, "%s", word_buffer);
        outputbuffer = AddToString(outputbuffer, text_to_print);
        j = 0;
      }
    i++;
    }
    else { /* (char) -1 */
      /* An id is coming next.                                   */
      /* Id syntax: -1 TAG MSB LSB (4 bytes/chars total length). */
      /* -1 indicates a parameter is coming.                     */
      /* TAG left nibble indicates whether MSB should be read as */
      /* a 0 (nibble is 1) or not (nibble is 2).                 */
      /* TAG right nibble indicates same as above for LSB.       */
      /* MSB is MSB value of parameter id.                       */
      /* LSB is LSB value of parameter id.                       */

      /* We must cast to unsigned char because we do not want to */
      /* use negative numbers in out math. If we change the      */
      /* type of str to unsigned, the check for -1 will fail.    */

      /* 22 jan 2017: we now also allow description ids and      */
      /* attribute ids. But these must have an owner. Owner is   */
      /* a location, an object or THIS keyword,                  */
      /* After a location id, an object id or a THIS keyword     */
      /* there will always follow a second id. In case this id   */
      /* ia a description or an attribute, the preceding id is   */
      /* its owner. If the second id is NO_ID, then the          */
      /* preceding id was a single occurrence and must be printed */

      id1  = ( (int32_t) (str[i+1] & 0xF0) ==  0x10? 0: (int32_t) (unsigned char) (str[i+2])*256);
      id1 += ( (int32_t) (str[i+1] & 0x0F) ==  0x01? 0: (int32_t) (unsigned char) str[i+3]);

      /* Skip the -1, the tag and the 2 bytes for the id.*/
      i += 4;

      /* Convert in case it is a special id. */
      ConvSpecId(&id1, &type);

      /* 02march2017: we must first check for a number */
      /* if we don't and it is a number with a value that    */
      /* is equal to a location id, object is, timer id, etc */
      /* things will go wrong.                               */
      if (type == NUMBER) {
        PrintNumber(id1);
      }
      else {
        /* not a number */
        if (IsLocId(id1) || IsObjId(id1)) {
          /* Another parameter is coming */
          /* loc/obj/THIS description    */
          /* loc/obj/THIS attribute      */
          /* loc/obj/THIS NO_ID          */

          id2  = ( (int32_t) (str[i+1] & 0xF0) ==  0x10? 0: (int32_t) (unsigned char) (str[i+2])*256);
          id2 += ( (int32_t) (str[i+1] & 0x0F) ==  0x01? 0: (int32_t) (unsigned char) str[i+3]);

          /* Skip the -1, the tag and the 2 bytes for the id.*/
          i += 4;
        }

        if (id1 == ARTICLE_A || id1 == ARTICLE_THE) {
          /* we must remember to print an article if the next */
          /* word is an object or location id                 */
          article = id1;
        }

        if (IsTimerId(id1)) {
          PrintId(id1);
          article = 0;
        }

        if (IsLocId(id1) || IsObjId(id1)) {
          if (id2 == NO_ID) {
            PrintId(id1);
            article = 0;
          }
          else {
            /* id2 must be a description or an attribute */
            /* now, id1 is the owner                     */
            /* Now set owner and id1 so we can use the   */
            /* same code for a single descr or attr      */
            owner = id1;
            id1   = id2;
          }
        }

        if (IsCAttrId(id1) || IsLAttrId(id1)) {
          if (!GetAttributeInfo(id1, owner, &attributes, &attribute_index)) {
            PrintError(48, NULL, NULL);
            return;
          }
          else {
            switch (attributes[attribute_index].type) {
              case ATTR_ID:
                PrintError(49, NULL, NULL);
                break;
              case COMMON_DESCRS: ;  /* common description */
              case DESCR_ID:         /* local description */
                id1 = attributes[attribute_index].value;
                /* Owner is the owner of the attribute, but */
                /* since the value is a description id we   */
                /* must also retrieve the description owner */
                /* from the attributes struct.              */
                owner = attributes[attribute_index].value_owner;
                break;
              case NUMBER:
                PrintNumber(attributes[attribute_index].value);
                break;
              default:
                if (attributes[attribute_index].value != NONE) {
                  PrintId(attributes[attribute_index].value);
                  article = 0;
                }
                else {
                  /* the attribute value is NONE, don't print anything  */
                  /* we must check if a white space was already printed */
                  /* if so, the space was already added to outputbuffer */
                  if (outputbuffer[strlen(outputbuffer)-1] == ' ') {
                    outputbuffer[strlen(outputbuffer)-1] = '\0';
                  }
                  /* NOTE: outputbuffer now ends with two '\0' chars */
                  /* THIS SHOULD NOT BE A PROBLEM                    */
                }
                break;
            }
          }
        }
        if (IsCDescrId(id1) || IsLDescrId(id1)) {
          /* 05feb2017: forgot to test for NULL pointer. Crashed. */
          if ((descr = GetDescr(owner, id1)) != NULL) {
            PrintString(descr);
          }
        }
        if (IsWordId(id1)) {
          PrintWord(id1);
        }
      } /* else for NUMBER test */
    }
  } /* while */
}


void PrintWord(int32_t id)
{
  wordTable *wt = word_table;
  int32_t   i   = 0;

  /* in case of NONE we must print a white space,  */
  /* except when we are at the beginning of a line */

  /* Find id. */
  while (i<nr_of_words && wt[i].id != id)
    i++;

  if (i == nr_of_words) {
    /* Internal error; this shouldn't happen. */
    PrintError(50, &((resultStruct) {VALUE, NONE, id}), NULL);
  }
  else
    PrintString(wt[i].print_word);
}



void PrintExtendedSysDescr(extendedSysDescr *extended_sys_descr)
{
  if (extended_sys_descr->dynamic != NULL) {
    if (!ConvertDynamicDSys(extended_sys_descr->dynamic, extended_sys_descr)) {
      return;
    }
  }

  PrintSysDescr(&(extended_sys_descr->part1));

  if (extended_sys_descr->connect_prepos != NO_ID) {
    PrintString(" ");
    PrintWord(extended_sys_descr->connect_prepos);
    PrintString(" ");

    /* PrintSysDescr() will not print the article for  */
    /* the second part of extended system descriptions */
    /* Because variable 'article' will be 0.           */
    if ((extended_sys_descr->part2).article != NO_ID) {
      PrintWord((extended_sys_descr->part2).article);
      PrintString(" ");
    }
    PrintSysDescr(&(extended_sys_descr->part2));
  }
  return;
}


void PrintSysDescr(sysDescr *descr)
{
  int32_t i = 0;

  /* Caller must print the leading space, if needed.           */
  /* Since we have a recursive syntax check, the adjectives    */
  /* have been inserted backwards (last one first). Therefore, */
  /* we will print the adjectives in reverse order.            */

  /* print the article (or not) */
  PrintArticle(descr, article);

  for (i=descr->nr_of_adjectives-1; i>=0; i--) {
    PrintWord(descr->adjectives[i]);

    /* clear article, otherwise the white space */
    /* will not be printed by PrintString()     */
    article = 0;
    PrintString(" ");
  }

  if (descr->noun != NO_ID)
    PrintWord(descr->noun);
}


void PrintArticle(sysDescr *descr, int32_t type)
{
  switch (story_info.story_language) {
    case NL:
      NL_PrintArticle(descr, type);
      break;
    case ENG:
      ENG_PrintArticle(descr, type);
      break;
    default:
      /* unknown language, use English */
      ENG_PrintArticle(descr, type);
      break;
  }
}


void PrintId(int32_t id)
{
  int32_t offset;
  int32_t type = NO_TYPE;
  int32_t num = 0;
  char    text_to_print[OUTPUT_LINE_LEN];

  /* Remember whether it's a number. */
  if (id == VALUE || id == ORDINAL)
    num = 1;

  /* Convert in case it is a special id. */
  ConvSpecId(&id, &type);

  if (num)
    PrintNumber(id);
  else if (IsLocId(id)) {
    offset = id - FIRST_LOCATION_ID;
    if (loc_dir[offset].nr_of_dsys != 0)
      PrintExtendedSysDescr(&(loc_dir[offset].descr[0]));
    /* else do nothing */
  }
  else if (IsObjId(id)) {
    offset = id - FIRST_OBJECT_ID;
    if (obj_dir[offset].nr_of_dsys != 0)
      PrintExtendedSysDescr(&(obj_dir[offset].descr[0]));
    /* else do nothing */
  }
  else if (IsTimerId(id)) {
    offset = id - FIRST_TIMER_ID;
    sprintf(text_to_print, "%d", timers[offset].value);
    PrintString(text_to_print);
  }
  else
    /* Assume it's a word id. */
    PrintWord(id);
}


void PrintNumber(int32_t number)
{
  char text_to_print[OUTPUT_LINE_LEN];

  if (muted == 1) {
    return;
  }

  sprintf(text_to_print, "%d", number);
  PrintString(text_to_print);
}


int32_t IsSpecialChar(char ch)
{
  if ( (((int32_t) ch>=97) && ((int32_t) ch<=122)) ||   /* a..z */
       (((int32_t) ch>=65) && ((int32_t) ch<=90))  ||   /* A..Z */
       (((int32_t) ch>=48) && ((int32_t) ch<=57))  ||   /* 0..9 */
       (ch == '_') )                          /* for identifiers */
    return(0);
  else
    return(1);
}


int32_t IsVowel(char ch)
{
  switch (ch) {
    case 'a':
    case 'A':
    case 'e':
    case 'E':
    case 'i':
    case 'I':
    case 'o':
    case 'O':
    case 'u':
    case 'U':
    case 'y':
    case 'Y':
      return(OK);
      break;
    default:
      return(ERROR);
  }
}


int32_t Power(int32_t i, int32_t j)
{
  /* Calculates i raised to the power of j. */
  /* Works only for j>0.                    */

  if (j == 0)
    return(1);
  else
    return(i*Power(i, j-1));
}
