
/************************************************************************/
/* Copyright (c) 2016 - 2022 Marnix van den Bos.                        */
/*                                                                      */
/* <marnix@xvan.nl>                                                     */
/*                                                                      */
/* This file is part of XVAN, an Interactive Fiction authoring system.  */
/*                                                                      */
/* XVAN is free software: you can redistribute it and/or modify         */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, version 3.                             */
/*                                                                      */
/* XVAN is distributed in the hope that it will be useful,              */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */
/*                                                                      */
/* You should have received a copy of the GNU General Public License    */
/* along with XVAN.  If not, see <http://www.gnu.org/licenses/>.        */
/*                                                                      */
/************************************************************************/


#include <stdio.h>
#include <stdlib.h>  /* malloc(). realloc(), free() */
#include <string.h>
#include <stdint.h>
#include "keyword.h"
#include "typedefs.h"
#include "restore.h"

/*************************/
/* function declarations */
/*************************/

int32_t ReadInt16(int16_t*, FILE*);
int32_t ReadInt32(int32_t*, FILE*);
int32_t Readint64(int64_t*, FILE*);
int32_t RestoreString(char*, FILE*);
int32_t RestoreSpecialIds(FILE*);
int32_t ReadId(int32_t*, FILE*);
int32_t RestoreStoryInfo(storyInfo*, FILE*);
int32_t RestoreExtendedSysDescr(extendedSysDescr*, FILE*);
int32_t RestoreSysDescr(sysDescr*, FILE*);
int32_t RestoreContData(contData*, FILE*);
int32_t RestoreDirInfo(dirInfo*, FILE*);
int32_t RestoreLocationDirectory(FILE*);
int32_t RestoreObjectDirectory(FILE*);
int32_t RestoreExits(FILE*);
int32_t ReadAttribute(attrInfo*, FILE*);
int32_t RestoreCommonAttributes(FILE*);
int32_t RestoreLocalAttributes(FILE*);
int32_t RestoreCommonFlags(FILE*);
int32_t RestoreLocalFlags(FILE*);
int32_t ReadTimer(timerInfo*, FILE*);
int32_t RestoreTimers(FILE*);

/************************/
/* function definitions */
/************************/

int32_t ReadInt16(int16_t *n, FILE *savefile)
{
  if (fread((void *) n, sizeof(int16_t), 1, savefile) != 1)
    /* error message will be printed by calling function */
    return(ERROR);

  /* convert to host endian format */
  *n = xv_ntohs(*n);

  return(OK);
}


int32_t ReadInt32(int32_t *n, FILE *savefile)
{
  if (fread((void *) n, sizeof(int32_t), 1, savefile) != 1)
    /* error message will be printed by calling function */
    return(ERROR);

  /* convert to host endian format */
  *n = xv_ntohl(*n);

  return(OK);
}


int32_t ReadInt64(int64_t *n, FILE *savefile)
{
  if (fread((void *) n, sizeof(int64_t), 1, savefile) != 1)
    /* error message will be printed by calling function */
    return(ERROR);

  /* convert to host endian format */
  *n = xv_ntohll(*n);

  return(OK);
}


int32_t RestoreString(char *target, FILE *savefile)
{
  int32_t len;

  /* read the length of the string */
  if (!ReadInt32(&len, savefile)) {
    /* error message will be printed by calling function */
    return(ERROR);
  }

  /* now read the string itself */
  if (fread((void *) target, sizeof(char), len, savefile) != len) {
    /* error message will be printed by calling function */
    return(ERROR);
  }

  /* add a '\0' (there will be room bacause it was saved succesfully */
  target[len] = '\0';

  return(OK);
}


int32_t ReadId(int32_t *id, FILE *savefile)
{
  if (!ReadInt32(id, savefile)) {
    PrintError(14, NULL, "ReadId()");
    return(ERROR);
  }
  return(OK);
}


int32_t RestoreSpecialIds(FILE *savefile)
{
  int32_t special_id_array[NR_OF_SPECIAL_IDS];
  int i = 0;

  /* read the special ids */
  for (i=0; i< 10; i++) {
    if (!ReadInt32(special_id_array+i, savefile)) {
      PrintError(14, NULL, "RestoreSpecialIds()");
      return(ERROR);
    }
  }
  /* now assign the special ids from the array */

  action        = special_id_array[0];
  prepos        = special_id_array[1];
  direction     = special_id_array[2];
  active_entity = special_id_array[3];
  curr_loc      = special_id_array[4];
  actor         = special_id_array[5];
  subject       = special_id_array[6];
  specifier     = special_id_array[7];
  value         = special_id_array[8];
  ordinal       = special_id_array[9];

  return(OK);
}


int32_t RestoreExtendedSysDescr(extendedSysDescr *extended_system_description, FILE *savefile)
{
  int32_t code = NO_ID;

  if (!ReadInt32(&code, savefile)) {
    PrintError(14, NULL, "RestoreExtendedSysDescr()");
    return(ERROR);
  }

  if (code == DYN_DSYS) {
    extended_system_description->dynamic = ReadString();
    return(OK);
  }

  /* it's a 'normal' system description */

  if (!RestoreSysDescr(&(extended_system_description->part1), savefile))
    return(ERROR);

  if (!ReadInt32(&(extended_system_description->connect_prepos), savefile)) {
    PrintError(14, NULL, "RestoreExtendedSysDescr()");
    return(ERROR);
  }

  if (!RestoreSysDescr(&(extended_system_description->part2), savefile))
    return(ERROR);

  return(OK);
}


int32_t RestoreSysDescr(sysDescr *system_description, FILE *savefile)
{
  int32_t i=0;

  if (!ReadInt32(&(system_description->article), savefile)) {
    PrintError(14, NULL, "RestoreSysDescr()");
    return(ERROR);
  }

  if (!ReadInt32(&(system_description->nr_of_adjectives), savefile)) {
    PrintError(14, NULL, "RestoreSysDescr()");
    return(ERROR);
  }

  for (i=0; i<MAX_PARSE_ADJ; i++)
    if (!ReadId((system_description->adjectives)+i, savefile)) {
      PrintError(14, NULL, "RestoreSysDescr()");
      return(ERROR);
    }

  if (!ReadId(&(system_description->noun), savefile)) {
    PrintError(14, NULL, "RestoreSysDescr()");
    return(ERROR);
  }

  return(OK);
}


int32_t RestoreContData(contData *cont_data, FILE *savefile)
{
  int32_t i=0;

  if (!ReadInt32(&(cont_data->nr_of_objects), savefile)) {
    PrintError(14, NULL, "RestoreContData()");
    return(ERROR);
  }

  for (i=0; i<MAX_CONTAINED_OBJECTS; i++)
    if (!ReadId((cont_data->object_ids)+i, savefile)) {
      PrintError(14, NULL, "RestoreContData()");
      return(ERROR);
    }

  return(OK);
}


int32_t RestoreStoryInfo(storyInfo *story_info, FILE *savefile)
{
  /* Read the storyInfo struct. */
  if (!RestoreString(story_info->title, savefile)) {
    PrintError(14, NULL, "RestoreStoryInfo()");
    return(ERROR);
  }

  if (!RestoreString(story_info->author, savefile)) {
    PrintError(14, NULL, "RestoreStoryInfo()");
    return(ERROR);
  }

  if (!RestoreString(story_info->organization, savefile)) {
    PrintError(56, NULL, "SaveStoryInfo()");
    return(ERROR);
  }

  if (!RestoreString(story_info->cover_text, savefile)) {
    PrintError(14, NULL, "RestoreStoryInfo()");
    return(ERROR);
  }

  if (!RestoreString(story_info->credits, savefile)) {
    PrintError(14, NULL, "RestoreStoryInfo()");
    return(ERROR);
  }

  if (!RestoreString(story_info->version, savefile)) {
    PrintError(14, NULL, "RestoreStoryInfo()");
    return(ERROR);
  }

  if (!RestoreString(story_info->android_mkt, savefile)) {
    PrintError(14, NULL, "RestoreStoryInfo()");
    return(ERROR);
  }

  if (!RestoreString(story_info->ios_mkt, savefile)) {
    PrintError(14, NULL, "RestoreStoryInfo()");
    return(ERROR);
  }

  if (!RestoreString(story_info->backimage, savefile)) {
    PrintError(14, NULL, "RestoreStoryInfo()");
    return(ERROR);
  }

  if (!RestoreString(story_info->effect, savefile)) {
    PrintError(14, NULL, "RestoreStoryInfo()");
    return(ERROR);
  }

  if (!RestoreString(story_info->primary_color, savefile)) {
    PrintError(14, NULL, "RestoreStoryInfo()");
    return(ERROR);
  }

  if (!ReadInt16(&(story_info->ui_sidebar), savefile)) {
    PrintError(14, NULL, "RestoreStoryInfo()");
    return(ERROR);
  }

  if (!ReadInt16(&(story_info->ui_textinput), savefile)) {
    PrintError(14, NULL, "RestoreStoryInfo()");
    return(ERROR);
  }

  if (!ReadInt16(&(story_info->ui_compass), savefile)) {
    PrintError(14, NULL, "RestoreStoryInfo()");
    return(ERROR);
  }

  if (!ReadInt16(&(story_info->autolink), savefile)) {
    PrintError(14, NULL, "RestoreStoryInfo()");
    return(ERROR);
  }

  if (!RestoreString(story_info->compiler_version, savefile)) {
    PrintError(56, NULL, "SaveStoryInfo()");
    return(ERROR);
  }

  if (!ReadInt16(&(story_info->xvan_language), savefile)) {
    PrintError(14, NULL, "RestoreStoryInfo()");
    return(ERROR);
  }

  if (!ReadInt16(&(story_info->story_language), savefile)) {
    PrintError(14, NULL, "RestoreStoryInfo()");
    return(ERROR);
  }

  if (!ReadInt16(&(story_info->play_mode), savefile)) {
    PrintError(14, NULL, "RestoreStoryInfo()");
    return(ERROR);
  }

  return(OK);
}


int32_t RestoreDirInfo(dirInfo *dir_info, FILE *savefile)
{
  int i = 0;

  /* read nr_of_dsys */
  if (!ReadInt32(&(dir_info->nr_of_dsys), savefile)) {
    PrintError(14, NULL, "RestoreDirInfo()");
    return(ERROR);
  }

  /* read dsys instances */
  for (i=0; i<MAX_DSYS; i++) {
    if (!RestoreExtendedSysDescr((dir_info->descr)+i, savefile))
      return(ERROR);
  }

  /* read contained objects */
  if (!RestoreContData(&(dir_info->contained_objs), savefile))
      return(ERROR);

  /* read containing object or location */
  if (!ReadInt32(&(dir_info->held_by), savefile)) {
    PrintError(14, NULL, "RestoreDirInfo()");
    return(ERROR);
  }

  /* read offset */
  if (!ReadInt64(&(dir_info->offset), savefile)) {
    PrintError(14, NULL, "RestoreDirInfo()");
    return(ERROR);
  }

  return(OK);
}


int32_t RestoreLocationDirectory(FILE *savefile)
{
  int32_t i = 0;

  for (i=0; i<nr_of_locs; i++) {
    if (!RestoreDirInfo(loc_dir+i, savefile))
      return(ERROR);
  }

  return(OK);
}


int32_t RestoreObjectDirectory(FILE *savefile)
{
  int32_t i=0;

  for (i=0; i<nr_of_objs; i++) {
    if (!RestoreDirInfo(obj_dir+i, savefile))
      return(ERROR);
  }
  return(OK);
}


int32_t RestoreExits(FILE *savefile)
{
  int32_t size; /* size of the exit_data array */
  int     i = 0;

  size = nr_of_locs * nr_of_directions;

  for (i=0; i<size; i++) {
    if (!ReadInt32(exit_data+i, savefile)) {
      PrintError(14, NULL, "RestoreExits()");
      return(ERROR);
    }
  }

  return(OK);
}


int32_t ReadAttribute(attrInfo *attribute, FILE *savefile)
{
  if (!ReadInt32(&(attribute->type), savefile)) {
    PrintError(14, NULL, "ReadAttribute()");
    return(ERROR);
  }

  if (!ReadInt32(&(attribute->value), savefile)) {
    PrintError(14, NULL, "ReadAttribute()");
    return(ERROR);
  }

  if (!ReadInt32(&(attribute->value_owner), savefile)) {
    PrintError(14, NULL, "ReadAttribute()");
    return(ERROR);
  }

  return(OK);
}


int32_t RestoreCommonAttributes(FILE *savefile)
{
  int32_t total_attrs; /* Total number of common attributes for all locations */
  int i = 0;

  /* Store the common location attributes */
  total_attrs = nr_of_cattrs*nr_of_locs;

  for (i=0; i<total_attrs; i++) {
    if (!ReadAttribute(c_loc_attrs+i, savefile))
      return(ERROR);
  }

  /* Store the common object attributes */
  total_attrs = nr_of_cattrs*nr_of_objs;

  for (i=0; i<total_attrs; i++) {
    if (!ReadAttribute(c_obj_attrs+i, savefile))
      return(ERROR);
  }

  return(OK);
}


int32_t RestoreLocalAttributes(FILE *savefile)
{
  int i = 0;
  /* Total number of local attributes is stored in global variable nr_of_lattrs */

  for (i=0; i<nr_of_lattrs; i++) {
    if (!ReadAttribute(local_attrs+i, savefile))
      return(ERROR);
  }

  return(OK);
}


int32_t RestoreCommonFlags(FILE *savefile)
{
  int32_t com_loc_flags_string_len = 0;
  int32_t com_obj_flags_string_len = 0;
  int i = 0;

  /* calculate the length of the common flag strings */
  com_loc_flags_string_len = ((nr_of_cflags*nr_of_locs)/WORD_LEN)+1;
  com_obj_flags_string_len = ((nr_of_cflags*nr_of_objs)/WORD_LEN)+1;

  for (i=0; i<com_loc_flags_string_len; i++) {
    if (!ReadInt32(com_loc_flags+i, savefile)) {
      PrintError(14, NULL, "RestoreCommonFlags()");
      return(ERROR);
    }
  }

  for (i=0; i<com_obj_flags_string_len; i++) {
    if (!ReadInt32(com_obj_flags+i, savefile)) {
      PrintError(14, NULL, "RestoreCommonFlags()");
      return(ERROR);
    }
  }

  return(OK);
}


int32_t RestoreLocalFlags(FILE *savefile)
{
  int i = 0;

  /* Global var loc_flags_string_len is used here */
  for (i=0; i<loc_flags_string_len; i++) {
    if (!ReadInt32(local_flags+i, savefile)) {
      PrintError(14, NULL, "RestoreLocalFlags()");
      return(ERROR);
    }
  }

  return(OK);
}


int32_t ReadTimer(timerInfo *timer, FILE *savefile)
{
  if (!ReadInt32(&(timer->value), savefile)) {
    PrintError(14, NULL, "ReadTimer()");
    return(ERROR);
  }
  if (!ReadInt32(&(timer->step), savefile)) {
    PrintError(14, NULL, "ReadTimer()");
    return(ERROR);
  }
  if (!ReadInt32(&(timer->interval), savefile)) {
    PrintError(14, NULL, "ReadTimer()");
    return(ERROR);
  }
  if (!ReadInt32(&(timer->update), savefile)) {
    PrintError(14, NULL, "ReadTimer()");
    return(ERROR);
  }
  if (!ReadInt16(&(timer->direction), savefile)) {
    PrintError(14, NULL, "ReadTimer()");
    return(ERROR);
  }

  if (!ReadInt16(&(timer->state), savefile)) {
    PrintError(14, NULL, "ReadTimer()");
    return(ERROR);
  }

  if (!ReadInt32(&(timer->trigger_at), savefile)) {
    PrintError(14, NULL, "ReadTimer()");
    return(ERROR);
  }

  if (!ReadInt32(&(timer->trigger_spec), savefile)) {
    PrintError(14, NULL, "ReadTimer()");
    return(ERROR);
  }

  if (!ReadInt32(&((timer->execute)[0]), savefile)) {
    PrintError(14, NULL, "ReadTimer()");
    return(ERROR);
  }

  if (!ReadInt32((&(timer->execute[1])), savefile)) {
    PrintError(14, NULL, "ReadTimer()");
    return(ERROR);
  }

  return(OK);
}


int32_t RestoreTimers(FILE *savefile)
{
  int i = 0;
 /* global vars timerInfo *timers and int nr_of_timers are used here */

  for (i=0; i<nr_of_timers; i++) {
    if (!ReadTimer(timers+i, savefile))
      return(ERROR);
  }

  return(OK);
}
