
/************************************************************************/
/* Copyright (c) 2016 - 2022 Marnix van den Bos.                        */
/*                                                                      */
/* <marnix@xvan.nl>                                                     */
/*                                                                      */
/* This file is part of XVAN, an Interactive Fiction authoring system.  */
/*                                                                      */
/* XVAN is free software: you can redistribute it and/or modify         */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, version 3.                             */
/*                                                                      */
/* XVAN is distributed in the hope that it will be useful,              */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */
/*                                                                      */
/* You should have received a copy of the GNU General Public License    */
/* along with XVAN.  If not, see <http://www.gnu.org/licenses/>.        */
/*                                                                      */
/************************************************************************/


#include <stdio.h>
#include <stdlib.h>  /* malloc(). realloc(), free() */
#include <string.h>
#include <stdint.h>
#include "keyword.h"
#include "typedefs.h"
#include "save.h"

/*************************/
/* function declarations */
/*************************/

int32_t StoreInt16(int16_t, FILE*);
int32_t StoreInt32(int32_t, FILE*);
int32_t Storeint64(int64_t, FILE*);
int32_t StoreId(int32_t, FILE*);
int32_t SaveString(char*, FILE*);
int32_t SaveSpecialIds(FILE*);
int32_t SaveStoryInfo(FILE*);
int32_t SaveExtendedSysDescr(extendedSysDescr*, FILE*);
int32_t SaveSysDescr(sysDescr*, FILE*);
int32_t SaveContData(contData*, FILE*);
int32_t SaveDirInfo(dirInfo*, FILE*);
int32_t SaveLocationDirectory(FILE*);
int32_t SaveObjectDirectory(FILE*);
int32_t SaveExits(FILE*);
int32_t SaveAttribute(attrInfo*, FILE*);
int32_t SaveCommonAttributes(FILE*);
int32_t SaveLocalAttributes(FILE*);
int32_t SaveCommonFlags(FILE*);
int32_t SaveLocalFlags(FILE*);
int32_t SaveTimer(timerInfo*, FILE*);
int32_t SaveAllTimers(FILE*);

/************************/
/* function definitions */
/************************/

int32_t StoreInt16(int16_t n, FILE *savefile)
{
  /* convert to big endian */
  n = xv_htons(n);

  if (fwrite((void *) &n, sizeof(int16_t), 1, savefile) != 1)
    /* error message will be printed by calling function */
    return(ERROR);
  return(OK);
}


int32_t StoreInt32(int32_t n, FILE *savefile)
{
  /* convert to big endian */
  n = xv_htonl(n);

  if (fwrite((void *) &n, sizeof(int32_t), 1, savefile) != 1)
    /* error message will be printed by calling function */
    return(ERROR);
  return(OK);
}


int32_t StoreInt64(int64_t n, FILE *savefile)
{
  /* convert to big endian */
  n = xv_htonll(n);

  if (fwrite((void *) &n, sizeof(int64_t), 1, savefile) != 1)
    /* error message will be printed by calling function */
    return(ERROR);
  return(OK);
}


int32_t StoreId(int32_t id, FILE *savefile)
{
  if (!StoreInt32(id, savefile)) {
    PrintError(56, NULL, "StoreId()");
    return(ERROR);
  }
  return(OK);
}


int32_t SaveString(char *source, FILE *savefile)
{
  int32_t len;

  len = strlen(source);

  /* store the string length */
  if (!StoreInt32(len, savefile)) {
    /* error message will be printed by calling function */
    return(ERROR);
  }

  /* now store the string itself */
  if (fwrite((void *) source, 1, len, savefile) != len) {
    /* error message will be printed by calling function */
    return(ERROR);
  }

  return(OK);
}


int32_t SaveSpecialIds(FILE *savefile)
{
  int32_t special_id_array[NR_OF_SPECIAL_IDS];
  int i = 0;

  /* now store the special ids in the array so we can save them in the savefile */

  special_id_array[0] = action;
  special_id_array[1] = prepos;
  special_id_array[2] = direction;
  special_id_array[3] = active_entity;
  special_id_array[4] = curr_loc;
  special_id_array[5] = actor;
  special_id_array[6] = subject;
  special_id_array[7] = specifier;
  special_id_array[8] = value;
  special_id_array[9] = ordinal;

  /* save the special ids */
  for (i=0; i< 10; i++) {
    if (!StoreInt32(special_id_array[i], savefile)) {
      PrintError(56, NULL, "SaveSpecialIds()");
    return(ERROR);
    }
  }

  return(OK);
}


int32_t SaveStoryInfo(FILE *savefile)
{
  /* Store the storyInfo struct. */
  if (!SaveString(story_info.title, savefile)) {
    PrintError(56, NULL, "SaveStoryInfo()");
    return(ERROR);
  }

  if (!SaveString(story_info.author, savefile)) {
    PrintError(56, NULL, "SaveStoryInfo()");
    return(ERROR);
  }

  if (!SaveString(story_info.organization, savefile)) {
    PrintError(56, NULL, "SaveStoryInfo()");
    return(ERROR);
  }

  if (!SaveString(story_info.cover_text, savefile)) {
    PrintError(56, NULL, "SaveStoryInfo()");
    return(ERROR);
  }

  if (!SaveString(story_info.credits, savefile)) {
    PrintError(56, NULL, "SaveStoryInfo()");
    return(ERROR);
  }

  if (!SaveString(story_info.version, savefile)) {
    PrintError(56, NULL, "SaveStoryInfo()");
    return(ERROR);
  }

  if (!SaveString(story_info.android_mkt, savefile)) {
    PrintError(56, NULL, "SaveStoryInfo()");
    return(ERROR);
  }

  if (!SaveString(story_info.ios_mkt, savefile)) {
    PrintError(56, NULL, "SaveStoryInfo()");
    return(ERROR);
  }

  if (!SaveString(story_info.backimage, savefile)) {
    PrintError(56, NULL, "SaveStoryInfo()");
    return(ERROR);
  }

  if (!SaveString(story_info.effect, savefile)) {
    PrintError(56, NULL, "SaveStoryInfo()");
    return(ERROR);
  }

  if (!SaveString(story_info.primary_color, savefile)) {
    PrintError(56, NULL, "SaveStoryInfo()");
    return(ERROR);
  }

  if (!StoreInt16(story_info.ui_sidebar, savefile)) {
    PrintError(56, NULL, "SaveStoryInfo()");
    return(ERROR);
  }

  if (!StoreInt16(story_info.ui_textinput, savefile)) {
    PrintError(56, NULL, "SaveStoryInfo()");
    return(ERROR);
  }

  if (!StoreInt16(story_info.ui_compass, savefile)) {
    PrintError(56, NULL, "SaveStoryInfo()");
    return(ERROR);
  }

  if (!StoreInt16(story_info.autolink, savefile)) {
    PrintError(56, NULL, "SaveStoryInfo()");
    return(ERROR);
  }

  if (!SaveString(story_info.compiler_version, savefile)) {
    PrintError(56, NULL, "SaveStoryInfo()");
    return(ERROR);
  }

  if (!StoreInt16(story_info.xvan_language, savefile)) {
    PrintError(56, NULL, "SaveStoryInfo()");
    return(ERROR);
  }

    if (!StoreInt16(story_info.story_language, savefile)) {
    PrintError(56, NULL, "SaveStoryInfo()");
    return(ERROR);
  }

    if (!StoreInt16(story_info.play_mode, savefile)) {
    PrintError(56, NULL, "SaveStoryInfo()");
    return(ERROR);
  }

  return(OK);
}


int32_t SaveExtendedSysDescr(extendedSysDescr *extended_system_description, FILE *savefile)
{
  /* first check if it is a dynamic system description */
  if (extended_system_description->dynamic != NULL) {
    /* dynamic system description */
    /* store keyword */
    if (!StoreInt32(DYN_DSYS, savefile)) {
      PrintError(56, NULL, "SaveExtendedSystemDescription()");
      return(ERROR);
    }
    /* store the string */
    if (!SaveString(extended_system_description->dynamic, savefile)) {
      PrintError(56, NULL, "SaveExtendedSystemDescription()");
      return(ERROR);
    }
    else {
      return(OK);
    }
  }

  /* not a dynamic system description */
  /* store keyword */
  if (!StoreInt32(DSYS, savefile)) {
    PrintError(56, NULL, "SaveExtendedSystemDescription()");
    return(ERROR);
  }
  if (!SaveSysDescr(&(extended_system_description->part1), savefile))
    return(ERROR);

  if (!StoreInt32(extended_system_description->connect_prepos, savefile)) {
      PrintError(56, NULL, "SaveExtendedSystemDescription()");
      return(ERROR);
  }

  if (!SaveSysDescr(&(extended_system_description->part2), savefile))
    return(ERROR);

  return(OK);
}


int32_t SaveSysDescr(sysDescr *system_description, FILE *savefile)
{
  int32_t i=0;

  if (!StoreInt32(system_description->article, savefile)) {
    PrintError(56, NULL, "SaveSysDescr()");
    return(ERROR);
  }

  if (!StoreInt32(system_description->nr_of_adjectives, savefile)) {
    PrintError(56, NULL, "SaveSysDescr()");
    return(ERROR);
  }

  for (i=0; i<MAX_PARSE_ADJ; i++)
    if (!StoreId((system_description->adjectives)[i], savefile)) {
      PrintError(56, NULL, "SaveSysDescr()");
      return(ERROR);
    }

  if (!StoreId(system_description->noun, savefile)) {
    PrintError(56, NULL, "SaveSysDescr()");
    return(ERROR);
  }
  return(OK);
}


int32_t SaveContData(contData *cont_data, FILE *savefile)
{
  int32_t i=0;

  if (!StoreInt32(cont_data->nr_of_objects, savefile)) {
    PrintError(56, NULL, "SaveContData()");
    return(ERROR);
  }

  for (i=0; i<MAX_CONTAINED_OBJECTS; i++)
    if (!StoreId((cont_data->object_ids)[i], savefile)) {
      PrintError(56, NULL, "SaveContData()");
      return(ERROR);
    }
  return(OK);
}


int32_t SaveDirInfo(dirInfo *dir_info, FILE *savefile)
{
  int i = 0;

  /* store nr_of_dsys */
  if (!StoreInt32(dir_info->nr_of_dsys, savefile)) {
    PrintError(56, NULL, "SaveDirInfo()");
    return(ERROR);
  }

  /* store dsys instances */
  for (i=0; i<MAX_DSYS; i++) {
    if (!SaveExtendedSysDescr(dir_info->descr+i, savefile))
      return(ERROR);
  }

  /* store contained objects */
  if (!SaveContData(&(dir_info->contained_objs), savefile))
      return(ERROR);

  /* store containing object or location */
  if (!StoreInt32(dir_info->held_by, savefile)) {
    PrintError(56, NULL, "SaveDirInfo()");
    return(ERROR);
  }

  /* store offset */
  if (!StoreInt64(dir_info->offset, savefile)) {
    PrintError(56, NULL, "SaveDirInfo()");
    return(ERROR);
  }

  return(OK);
}


int32_t SaveLocationDirectory(FILE *savefile)
{
  int32_t i = 0;

  for (i=0; i<nr_of_locs; i++) {
    if (!SaveDirInfo(loc_dir+i, savefile))
      return(ERROR);
  }

  return(OK);
}


int32_t SaveObjectDirectory(FILE *savefile)
{
  int32_t i=0;

  for (i=0; i<nr_of_objs; i++) {
    if (!SaveDirInfo(obj_dir+i, savefile))
      return(ERROR);
  }
  return(OK);
}


int32_t SaveExits(FILE *savefile)
{
  int32_t size; /* size of the exit_data array */
  int     i = 0;

  size = nr_of_locs * nr_of_directions;

  for (i=0; i<size; i++) {
    if (!StoreInt32(exit_data[i], savefile)) {
      PrintError(56, NULL, "SaveExits()");
      return(ERROR);
    }
  }

  return(OK);
}


int32_t SaveAttribute(attrInfo *attribute, FILE *savefile)
{
  if (!StoreInt32(attribute->type, savefile)) {
    PrintError(56, NULL, "SaveAttribute()");
    return(ERROR);
  }

  if (!StoreInt32(attribute->value, savefile)) {
    PrintError(56, NULL, "SaveAttribute()");
    return(ERROR);
  }

  if (!StoreInt32(attribute->value_owner, savefile)) {
    PrintError(56, NULL, "SaveAttribute()");
    return(ERROR);
  }

  return(OK);
}


int32_t SaveCommonAttributes(FILE *savefile)
{
  int32_t total_attrs; /* Total number of common attributes for all locations */
  int i = 0;

  /* Store the common location attributes */
  total_attrs = nr_of_cattrs*nr_of_locs;

  for (i=0; i<total_attrs; i++) {
    if (!SaveAttribute(c_loc_attrs+i, savefile))
      return(ERROR);
  }

  /* Store the common object attributes */
  total_attrs = nr_of_cattrs*nr_of_objs;

  for (i=0; i<total_attrs; i++) {
    if (!SaveAttribute(c_obj_attrs+i, savefile))
      return(ERROR);
  }

  return(OK);
}


int32_t SaveLocalAttributes(FILE *savefile)
{
  int i = 0;
  /* Total number of local attributes is stored in global variable nr_of_lattrs */

  for (i=0; i<nr_of_lattrs; i++) {
    if (!SaveAttribute(local_attrs+i, savefile))
      return(ERROR);
  }

  return(OK);
}


int32_t SaveCommonFlags(FILE *savefile)
{
  int32_t com_loc_flags_string_len = 0;
  int32_t com_obj_flags_string_len = 0;
  int i = 0;

  /* calculate the length of the common flag strings */
  com_loc_flags_string_len = ((nr_of_cflags*nr_of_locs)/WORD_LEN)+1;
  com_obj_flags_string_len = ((nr_of_cflags*nr_of_objs)/WORD_LEN)+1;

  for (i=0; i<com_loc_flags_string_len; i++) {
    if (!StoreInt32(com_loc_flags[i], savefile)) {
      PrintError(56, NULL, "SaveCommonFlags()");
      return(ERROR);
    }
  }

  for (i=0; i<com_obj_flags_string_len; i++) {
    if (!StoreInt32(com_obj_flags[i], savefile)) {
      PrintError(56, NULL, "SaveCommonFlags()");
      return(ERROR);
    }
  }

  return(OK);
}


int32_t SaveLocalFlags(FILE *savefile)
{
  int i = 0;

  /* Global var loc_flags_string_len is used here */
  for (i=0; i<loc_flags_string_len; i++) {
    if (!StoreInt32(local_flags[i], savefile)) {
      PrintError(56, NULL, "SaveLocalFlags()");
      return(ERROR);
    }
  }

  return(OK);
}


int32_t SaveTimer(timerInfo *timer, FILE *savefile)
{
  if (!StoreInt32(timer->value, savefile)) {
    PrintError(56, NULL, "SaveTimer()");
    return(ERROR);
  }
  if (!StoreInt32(timer->step, savefile)) {
    PrintError(56, NULL, "SaveTimer()");
    return(ERROR);
  }
  if (!StoreInt32(timer->interval, savefile)) {
    PrintError(56, NULL, "SaveTimer()");
    return(ERROR);
  }
  if (!StoreInt32(timer->update, savefile)) {
    PrintError(56, NULL, "SaveTimer()");
    return(ERROR);
  }
  if (!StoreInt16(timer->direction, savefile)) {
    PrintError(56, NULL, "SaveTimer()");
    return(ERROR);
  }
  if (!StoreInt16(timer->state, savefile)) {
    PrintError(56, NULL, "SaveTimer()");
    return(ERROR);
  }
  if (!StoreInt32(timer->trigger_at, savefile)) {
    PrintError(56, NULL, "SaveTimer()");
    return(ERROR);
  }
  if (!StoreInt32(timer->trigger_spec, savefile)) {
    PrintError(56, NULL, "SaveTimer()");
    return(ERROR);
  }
  if (!StoreInt32((timer->execute)[0], savefile)) {
    PrintError(56, NULL, "SaveTimer()");
    return(ERROR);
  }
  if (!StoreInt32((timer->execute[1]), savefile)) {
    PrintError(56, NULL, "SaveTimer()");
    return(ERROR);
  }

  return(OK);
}


int32_t SaveAllTimers(FILE *savefile)
{
  int i = 0;
 /* global vars timerInfo *timers and int nr_of_timers are used here */

  for (i=0; i<nr_of_timers; i++) {
    if (!SaveTimer(timers+i, savefile))
      return(ERROR);
  }

  return(OK);
}
